/*
 * Tabs are 4 symbols wide due to stupid Windows function names...
 * ---------------------------------------------------------------
 *
 * Copyright (c) 2010 KSI Labs, LLC. All rights reserved.
 * Written by Sergey Kubushyn <ksi@koi8.net>, <ksi@ksilabs.com>.
 *
 * File: centipede.c
 *
 * This is a WDM driver for KSI Labs, LLC CENTIPEDE PCI Board. Actually
 * it is not a driver but kernel level Engine for ArtSoft Mach3 CNC software
 * but it is made as a driver. It only supports several IOCTLs, no classic
 * I/O, no Power management, no WMI, nothing. All interactions with Mach3
 * software is through its shared memory block (TrajBuffer structure.)
 *
 * It replaces original ArtSoft Parallel Port Pulser Engine and emulates
 * all (well, most of) its functionality and interface and extends it to
 * include all the CENTIPEDE functionality without any changes to the Mach3
 * software. It adds the CENTIPEDE board as Port #10 that can be configured
 * off of Mach3 "Ports & Pins" dialog. Some signals are fixed to particular
 * CENTIPEDE pins, others can be assigned as it fits customer requirements.
 *
 * Parallel Ports can be also used for Mach3 I/O along with CENTIPEDE board.
 * However motion and limit/home/probe switches MUST be assigned to dedicated
 * CENTIPEDE pins because their processing is mostly done in the hardware
 * unlike original Pulser Engine.
 *
 * Please read documentation or go to http://www.ksilabs.com for further info.
 */

/*
 * 						LICENSE and Terms of Use.
 *
 * First of all, this is NOT GPL or Free Software. It is open so you have
 * all the source code for the Engine. There is also VHDL source code for
 * CENTIPEDE board CPLD that is available from KSI Labs, LLC website,
 * http://www.ksilabs.com.
 *
 * As for this particular code, Windows Mach3 Engine, you are free to use it,
 * make any modifications to customize it for your own needs, fix inevitable
 * bugs, port it to other CNC software and do anything else as long as your
 * derivative work that is based on this code or uses parts or pieces thereof
 * is made to be used with any KSI Labs, LLC products.
 *
 * We do NOT grant you a permission to use any parts of this software in any
 * products that are designed for products not manufactured or distributed
 * by KSI Labs, LLC. We might give such a permission on case-by-case basis
 * but it is not legal to use this software or bits and pieces thereof with
 * products not coming from KSI Labs, LLC without our explicit written
 * permission.
 *
 * The original copyright notices must be included in all derivative works.
 * You do NOT have to ask our permission to make any modifications to this
 * code and you are not under obligation to give us any of your work back
 * as long as you are using it for your own needs.
 *
 * However we do NOT give our permission to distribute your derivative work
 * unless all original copyrights are retained, all the conditions of this
 * license are met (use only with KSI Labs, LLC products) and full source
 * code for such derivative work is available for free to everybody without
 * any further conditions including but not limited to signing any kinds of
 * NDA. We explicitely deny permission to distribute any derivative work
 * based on this code with closed source binary-only fragments or include
 * bits or pieces of this code in such products.
 *
 * CNC software like Mach3 and likes is excluded from these requirements as
 * long as it does not include parts or pieces of this software into the
 * close source proprietary application proper. It can use shared memory
 * buffer, IOCTL calls, etc. to communicate with this Engine or the one based
 * on it and keep all the software making those calls or working on that
 * buffer close source and proprietary as long as those calls/structures are
 * documented and that documentation is freely available to everybody without
 * any further conditions including but no limited to signing any kind of NDA.
 */

#include "precomp.h"
#include "mach3.h"

#define _DRIVER_NAME_ "CENTIPEDE"

NTSTATUS DispatchPower(PDEVICE_OBJECT DeviceObject, IN PIRP Irp);

#ifdef ALLOC_PRAGMA
#pragma alloc_text (INIT, DriverEntry)
#pragma alloc_text (PAGE, AddDevice)
#pragma alloc_text (PAGE, Create)
#pragma alloc_text (PAGE, Close)
//#pragma alloc_text (PAGE, DispatchPnp)
#pragma alloc_text (PAGE, SystemControl)
#pragma alloc_text (PAGE, StartDevice)
#pragma alloc_text (PAGE, Unload)
#pragma alloc_text (PAGE, SendIrpSynchronously)
#endif


// ***********************************************
// * This is main Drive Entry Point. Gets called *
// * by the system when driver is loaded.        *
// ***********************************************
NTSTATUS DriverEntry(IN PDRIVER_OBJECT DriverObject, IN PUNICODE_STRING RegistryPath)
{
#ifdef DEBUG
	DebugPrint(INFO, DBG_INIT, "Entered DriverEntry of "_DRIVER_NAME_" version built on "
											__DATE__" at "__TIME__ "\n");
#if defined(WIN2K)
	DebugPrint(INFO, DBG_INIT, "Built in the Win2K build environment\n");
#endif
#endif

	UNREFERENCED_PARAMETER(RegistryPath);

	//
	// We do NOT support I/O, only IOCTLs
	//
	DriverObject->MajorFunction[IRP_MJ_PNP]            = DispatchPnp;
	DriverObject->MajorFunction[IRP_MJ_POWER]          = DispatchPower;
	DriverObject->MajorFunction[IRP_MJ_CREATE]         = Create;
	DriverObject->MajorFunction[IRP_MJ_CLOSE]          = Close;
	DriverObject->MajorFunction[IRP_MJ_CLEANUP]        = Cleanup;
	DriverObject->MajorFunction[IRP_MJ_DEVICE_CONTROL] = DispatchIoctl;
	DriverObject->MajorFunction[IRP_MJ_SYSTEM_CONTROL] = SystemControl;
	DriverObject->DriverExtension->AddDevice           = AddDevice;
	DriverObject->DriverUnload                         = Unload;

	return STATUS_SUCCESS;
}


// ******************************************************************
// * This is called by PnP layer on each and every CENTIPEDE device *
// * discovered on PCI bus as long as it is the only one :) Mach3   *
// * does not need more than one board hence no support for that.   *
// ******************************************************************
NTSTATUS AddDevice(PDRIVER_OBJECT DriverObject, PDEVICE_OBJECT PhysicalDeviceObject)
{
	NTSTATUS		status = STATUS_SUCCESS;
	PDEVICE_OBJECT	deviceObject = NULL;
	PFDO_DATA		FdoData;
	POWER_STATE		powerState;
	static BOOLEAN	devAdded = FALSE;

	PAGED_CODE();

#ifdef DEBUG
	DebugPrint(INFO, DBG_PNP, "-->AddDevice() PDO (0x%p)\n", PhysicalDeviceObject);
#endif

	if (devAdded) {
#ifdef DEBUG
		DebugPrint(ERROR, DBG_PNP, "AddDevice(): Only one board supported.\n");
#endif
		return STATUS_NOT_IMPLEMENTED;
	}

	status = IoCreateDevice(DriverObject,
							sizeof(FDO_DATA),
							NULL,
							FILE_DEVICE_UNKNOWN,
							FILE_DEVICE_SECURE_OPEN,
							FALSE,
							&deviceObject);

	if (!NT_SUCCESS(status)) return status;

	FdoData = (PFDO_DATA)deviceObject->DeviceExtension;

	RtlZeroMemory(FdoData, sizeof(FDO_DATA));

	FdoData->Signature = CPDDRV_FDO_INSTANCE_SIGNATURE;

	INITIALIZE_PNP_STATE(FdoData);

	FdoData->UnderlyingPDO = PhysicalDeviceObject;

	FdoData->QueueState = HoldRequests;

	FdoData->Self = deviceObject;
	FdoData->NextLowerDriver = NULL;

	InitializeListHead(&FdoData->NewRequestsQueue);
	KeInitializeSpinLock(&FdoData->QueueLock);

	//
	// OutstandingIO count is biased to 2. It transitions to 1 if the device
	// needs to be temporarily stopped. It transitions to zero when the
	// device needs to be removed.
	//
	FdoData->OutstandingIO = 2;

	//
	// Initialize the remove event to Not-Signaled. This event will be
	// set when the OutstandingIO will become 0.
	//
	KeInitializeEvent(&FdoData->RemoveEvent, SynchronizationEvent, FALSE);
	//
	// Initialize the stop event to Not-Signaled. This event will be set
	// when the OutstandingIO will become 1.
	//
	KeInitializeEvent(&FdoData->StopEvent, SynchronizationEvent, FALSE);

	//deviceObject->Flags |= DO_POWER_PAGABLE;
	deviceObject->Flags |= DO_BUFFERED_IO;

	FdoData->DevicePowerState = PowerDeviceD3;
	FdoData->SystemPowerState = PowerSystemWorking;

	powerState.DeviceState = PowerDeviceD3;
	PoSetPowerState(deviceObject, DevicePowerState, powerState);

	FdoData->NextLowerDriver = IoAttachDeviceToDeviceStack(deviceObject,
													PhysicalDeviceObject);
	if (!FdoData->NextLowerDriver) {
		IoDeleteDevice(deviceObject);
		return STATUS_NO_SUCH_DEVICE;
	}

	status = IoRegisterDeviceInterface(PhysicalDeviceObject,
									(LPGUID) &GUID_DEVINTERFACE_CPDDRV,
									NULL,
									&FdoData->InterfaceName);

	if (!NT_SUCCESS(status)) {
#ifdef DEBUG
		DebugPrint(ERROR, DBG_PNP,
			"AddDevice(): IoRegisterDeviceInterface failed (%x)\n", status);
#endif
		IoDetachDevice(FdoData->NextLowerDriver);
		IoDeleteDevice(deviceObject);
		return status;
	}

	devAdded = TRUE;

	//
	// Initialize our generic SpinLock
	//
	KeInitializeSpinLock(&FdoData->Lock);

	deviceObject->Flags &= ~DO_DEVICE_INITIALIZING;

#ifdef DEBUG
	DebugPrint(INFO, DBG_PNP, "<--AddDevice()\n");
#endif

	return status;
}


// *******************************************************
// This one is called on WMI IRP. We do NOT support WMI. *
// WMI IRP is simply passed down the stack.              *
// *******************************************************
NTSTATUS SystemControl(PDEVICE_OBJECT DeviceObject, PIRP Irp)
{
	PFDO_DATA	FdoData = (PFDO_DATA)DeviceObject->DeviceExtension;

	PAGED_CODE();

#ifdef DEBUG
	DebugPrint(TRACE, DBG_WMI, "<-->SystemControl()\n");
#endif

	IoIncrement(FdoData);
	//
	// Do nothing, just pass it down -- no WMI is spoken here.
	//
	return ForwardAndForget(FdoData, Irp);
}


// *****************************************************
// * This is where all PnP IRPs are sent by the system *
// *****************************************************
NTSTATUS DispatchPnp(PDEVICE_OBJECT DeviceObject, PIRP Irp)
{
	PFDO_DATA               FdoData;
	PIO_STACK_LOCATION      stack;
	NTSTATUS                status = STATUS_SUCCESS;
	PPNP_DEVICE_STATE       deviceState;
	UINT					mapBufOn;

	PAGED_CODE();

	FdoData = (PFDO_DATA)DeviceObject->DeviceExtension;

	stack = IoGetCurrentIrpStackLocation(Irp);

#ifdef DEBUG
	DebugPrint(TRACE, DBG_PNP, "-->DispatchPNP() %s \n",
				PnPMinorFunctionString(stack->MinorFunction));
#endif

	IoIncrement(FdoData);

	if (FdoData->DevicePnPState == Deleted) {
		Irp->IoStatus.Status = STATUS_NO_SUCH_DEVICE;
		IoCompleteRequest(Irp, IO_NO_INCREMENT);
		IoDecrement(FdoData);
		return STATUS_NO_SUCH_DEVICE;
	}

	mapBufOn = 0;
	if (FdoData->MapBufIrp) mapBufOn++;
	if (FdoData->MapTDIrp) mapBufOn++;

	switch (stack->MinorFunction) {

		case IRP_MN_START_DEVICE:
			status = SendIrpSynchronously(FdoData->NextLowerDriver, Irp);

			if (NT_SUCCESS(status))
				status = StartDevice(FdoData, Irp);

			break;

		case IRP_MN_QUERY_STOP_DEVICE:
			//
			// OK, we CAN stop our device, no reason to check it.
			// First, don't allow any requests to be passed down.
			//
			SET_NEW_PNP_STATE(FdoData, StopPending);
			FdoData->QueueState = HoldRequests;

			//
			// Wait for the all other I/Os to finish. 1 represents
			// the current query-stop being processed. 2 represents
			// MapBuffer IOCTL that may be still pending (mapBufOn.)
			//
			ReleaseAndWait(FdoData, 1 + mapBufOn, DEV_STOP);

			Irp->IoStatus.Status = STATUS_SUCCESS;

			return ForwardAndForget(FdoData, Irp);

		case IRP_MN_CANCEL_STOP_DEVICE:
			//
			// Send this IRP down and wait for it to come back.
			// Set the QueueState flag to AllowRequests,
			// and process all the previously queued up IRPs.
			// Also check to see whether you have received cancel-stop
			// without first receiving a query-stop. This could happen if
			// someone above us fails a query-stop and passes down the
			// subsequent cancel-stop.
			//
			status = SendIrpSynchronously(FdoData->NextLowerDriver, Irp);

			if (NT_SUCCESS(status) && FdoData->DevicePnPState == StopPending) {

				RESTORE_PREVIOUS_PNP_STATE(FdoData);

				FdoData->QueueState = AllowRequests;

				ASSERT(FdoData->DevicePnPState == Started);

				ProcessQueuedRequests(FdoData);
			}

			break;

		case IRP_MN_STOP_DEVICE:
			//
			// After the stop IRP has been sent to the lower driver object,
			// the driver must not send any more IRPs down that touch
			// the device until another START has occurred. For this reason
			// we are holding IRPs and waiting for the outstanding requests
			// to be finished when QUERY_STOP is received. IRP_MN_STOP_DEVICE
			// doesn't change anything in this behavior (we continue to hold
			// IRPs until a IRP_MN_START_DEVICE is issued).

			SET_NEW_PNP_STATE(FdoData, Stopped);

			//
			// This is the right place to actually give up all the resources
			// used. This might include calls to IoDisconnectInterrupt,
			// MmUnmapIoSpace, etc.
			// However we keep the SOFT_*_BUFF IRPs pending if they were
			// pending at this moment. This is to keep the kernel space
			// mapping for Mach3 TrajBuffer and the buffer itself locked in
			// the system memory. This mapping is initiated by the
			// application and there is no way for the driver to reestablish
			// this if START_DEVICE is commanded later.
			//
			status = ReturnResources(DeviceObject);

			Irp->IoStatus.Status = STATUS_SUCCESS;
			return ForwardAndForget(FdoData, Irp);

		case IRP_MN_QUERY_REMOVE_DEVICE:

			SET_NEW_PNP_STATE(FdoData, RemovePending);

			FdoData->QueueState = HoldRequests;

			ReleaseAndWait(FdoData, 1 + mapBufOn, DEV_STOP);

			Irp->IoStatus.Status = STATUS_SUCCESS;

			return ForwardAndForget(FdoData, Irp);

		case IRP_MN_CANCEL_REMOVE_DEVICE:
			//
			// We need to reset the QueueState flag to ProcessRequest,
			// since the device resume its normal activities.
			// Also check to see whether you have received cancel-remove
			// without first receiving a query-remove. This could happen if
			// someone above us fails a query-remove and passes down the
			// subsequent cancel-remove.
			//
			status = SendIrpSynchronously(FdoData->NextLowerDriver,Irp);

			if (NT_SUCCESS(status) && FdoData->DevicePnPState == RemovePending) {

				RESTORE_PREVIOUS_PNP_STATE(FdoData);

				FdoData->QueueState = AllowRequests;

				//
				// Process the queued requests that arrived between
				// QUERY_REMOVE and CANCEL_REMOVE.
				//
				ProcessQueuedRequests(FdoData);
			}

			break;

		case IRP_MN_SURPRISE_REMOVAL:
			//
			// The device has been unexpectedly removed from the machine
			// and is no longer available for I/O ("surprise" removal).
			// We must return device and memory resources, disable interfaces.
			// We will defer failing any outstanding request to IRP_MN_REMOVE.
			//
			// Note: This should NEVER happen to us, we are not removable...

			SET_NEW_PNP_STATE(FdoData, SurpriseRemovePending);
			FdoData->QueueState = FailRequests;

			//
			// Fail all the pending requests. Since the QueueState is
			// FailRequests ProcessQueuedRequests will simply flush
			// the queue, completing each IRP with STATUS_NO_SUCH_DEVICE
			//
			ProcessQueuedRequests(FdoData);

			//
			// Disable the device interface.
			//
			status = IoSetDeviceInterfaceState(&FdoData->InterfaceName, FALSE);

#ifdef DEBUG
			if (!NT_SUCCESS(status))
				DebugPrint(ERROR, DBG_PNP,
						"IoSetDeviceInterfaceState failed: 0x%x\n", status);
#endif
			//
			// Return any resources acquired during device startup.
			//
			ReturnResources(DeviceObject);

			//
			// We must set Irp->IoStatus.Status to STATUS_SUCCESS before
			// passing it down.
			//
			Irp->IoStatus.Status = STATUS_SUCCESS;

			return ForwardAndForget(FdoData, Irp);

		case IRP_MN_REMOVE_DEVICE:
			//
			// The Plug & Play system has dictated the removal of this device.
			// We have no choice but to detach and delete the device object.
			// (If we wanted to express an interest in preventing this removal,
			// we should have failed the query remove IRP).
			//

			SET_NEW_PNP_STATE(FdoData, Deleted);

			if (FdoData->PreviousPnPState != SurpriseRemovePending) {
				//
				// This means we are here after query-remove.
				// So first stop the device, disable the interface,
				// return resources, and fail all the pending requests.
				//
				FdoData->QueueState = FailRequests;

				//
				// Fail all the pending requests. Since the QueueState is
				// FailRequests ProcessQueuedRequests will simply flush
				// the queue, completing each IRP with STATUS_NO_SUCH_DEVICE
				//
				ProcessQueuedRequests(FdoData);

				//
				// Disable the Interface
				//
				status = IoSetDeviceInterfaceState(&FdoData->InterfaceName, FALSE);

#ifdef DEBUG
				if (!NT_SUCCESS(status))
					DebugPrint(ERROR, DBG_PNP,
						   "IoSetDeviceInterfaceState failed: 0x%x\n", status);
#endif
				//
				// Return hardware resources.
				//
				ReturnResources(DeviceObject);
			}

			//
			// Wait for all other I/O to finish. 1 represents
			// the current Remove IRP being processed.
			//
			ReleaseAndWait(FdoData, 1 + mapBufOn, DEV_REMOVE);

			//
			// Complete SOFT_ALLOCATE_BUFF and SOFT_SAMPLE_BUFF IRPs if they
			// are still pending
			//
			if (FdoData->MapBufIrp) {
				if (IoSetCancelRoutine(FdoData->MapBufIrp, NULL)) {
					FdoData->MapBufIrp->IoStatus.Information = 0;
					FdoData->MapBufIrp->IoStatus.Status = STATUS_CANCELLED;
					IoCompleteRequest(FdoData->MapBufIrp, IO_NO_INCREMENT);
					IoDecrement(FdoData);
					//
					// This is not needed because we are destroying the entire
					// DeviceObject but it won't do any harm...
					//
					FdoData->MapBufIrp = NULL;
					FdoData->MachSharedBlock = NULL;
				}
			}
			if (FdoData->MapTDIrp) {
				if (IoSetCancelRoutine(FdoData->MapTDIrp, NULL)) {
					FdoData->MapTDIrp->IoStatus.Information = 0;
					FdoData->MapTDIrp->IoStatus.Status = STATUS_CANCELLED;
					IoCompleteRequest(FdoData->MapTDIrp, IO_NO_INCREMENT);
					IoDecrement(FdoData);
					//
					// This is not needed because we are destroying the entire
					// DeviceObject but it won't do any harm...
					//
					FdoData->MapTDIrp = NULL;
					FdoData->MachTDBlock = NULL;
				}
			}

			//
			// Send on the remove IRP.
			// We need to send the remove down the stack before we detach,
			// but we don't need to wait for the completion of this operation
			// (and to register a completion routine).
			//
			Irp->IoStatus.Status = STATUS_SUCCESS;
			IoSkipCurrentIrpStackLocation(Irp);
			status = IoCallDriver(FdoData->NextLowerDriver, Irp);

			//
			// Detach the FDO from the device stack
			//
			IoDetachDevice(FdoData->NextLowerDriver);

			//
			// Free up interface memory
			//
			RtlFreeUnicodeString(&FdoData->InterfaceName);

			IoDeleteDevice (FdoData->Self);

			return status;

		default:
			//
			// Pass down all the unhandled Irps.
			//
			return ForwardAndForget(FdoData, Irp);
	}

	Irp->IoStatus.Status = status;
	IoCompleteRequest(Irp, IO_NO_INCREMENT);

	IoDecrement(FdoData);

#ifdef DEBUG
	DebugPrint(TRACE, DBG_PNP, "<--DispatchPNP()\n");
#endif

	return status;
}


// **********************************************************
// * Completion routine for PnP IRPs. Used in several other *
// * routines. E.g. SendIrpSynchronously() uses it.         *
// **********************************************************
NTSTATUS DispatchPnpComplete(PDEVICE_OBJECT DeviceObject, PIRP Irp, PVOID Context)
{
	PKEVENT	event = (PKEVENT)Context;

	UNREFERENCED_PARAMETER(DeviceObject);

#ifdef DEBUG
	DebugPrint(TRACE, DBG_PNP, "-->DispatchPNPComplete()\n");
#endif

	//
	// If the lower driver didn't return STATUS_PENDING, we don't need to
	// set the event because we won't be waiting on it. This optimization
	// avoids grabbing the dispatcher lock and improves perf.
	//
	if (Irp->PendingReturned)
		KeSetEvent(event, IO_NO_INCREMENT, FALSE);

#ifdef DEBUG
	DebugPrint(TRACE, DBG_PNP, "<--DispatchPNPComplete()\n");
#endif

	//
	// The dispatch routine must call IoCompleteRequest
	//
	return STATUS_MORE_PROCESSING_REQUIRED;
}


// ****************************************************
// * This is kinda open() routine that is called when *
// * user app calls CreateFile() on the device.       *
// ****************************************************
NTSTATUS Create(PDEVICE_OBJECT DeviceObject, PIRP Irp)
{
	PFDO_DATA	FdoData;

	PAGED_CODE();

	FdoData = (PFDO_DATA)DeviceObject->DeviceExtension;

#ifdef DEBUG
	DebugPrint(TRACE, DBG_CREATE_CLOSE, "-->Create()\n");
#endif

	//
	// Since we don't access the hardware to process create, we don't
	// have to worry about the current device state and whether to
	// queue this request or not.
	//

	IoIncrement(FdoData);

	if (FdoData->DevicePnPState == Deleted) {
		//
		// Since the device is removed, we will not hold any IRPs.
		// We just fail it.
		//
		Irp->IoStatus.Status = STATUS_NO_SUCH_DEVICE;
		IoCompleteRequest(Irp, IO_NO_INCREMENT);
		IoDecrement(FdoData);
		return STATUS_NO_SUCH_DEVICE;
	}

	Irp->IoStatus.Information = 0;
	Irp->IoStatus.Status = STATUS_SUCCESS;
	IoCompleteRequest(Irp, IO_NO_INCREMENT);

	IoDecrement(FdoData);

#ifdef DEBUG
	DebugPrint(TRACE, DBG_CREATE_CLOSE, "<--Create()\n");
#endif

	return STATUS_SUCCESS;
}


// *************************************************
// * This one gets called on Close() from user app *
// *************************************************
NTSTATUS Close(PDEVICE_OBJECT DeviceObject, PIRP Irp)
{
	PFDO_DATA	FdoData;
	NTSTATUS	status;

	PAGED_CODE();

	FdoData = (PFDO_DATA)DeviceObject->DeviceExtension;

#ifdef DEBUG
	DebugPrint(TRACE, DBG_CREATE_CLOSE, "-->Close()\n");
#endif

	IoIncrement(FdoData);

	status = STATUS_SUCCESS;

	Irp->IoStatus.Information = 0;
	Irp->IoStatus.Status = status;
	IoCompleteRequest(Irp, IO_NO_INCREMENT);

	IoDecrement(FdoData);

#ifdef DEBUG
	DebugPrint(TRACE, DBG_CREATE_CLOSE, "<--Close()\n");
#endif

	return status;
}


// ***********************************************************
// * Performs whatever initialization is needed to setup the *
// * device, namely connecting to an interrupt, setting up a *
// * DMA channel or mapping any I/O port resources.          *
// ***********************************************************
NTSTATUS StartDevice(PFDO_DATA FdoData, PIRP Irp)
{
	NTSTATUS            status = STATUS_SUCCESS;
	PIO_STACK_LOCATION  stack;
	UINT32				tmp;
	POWER_STATE         powerState;

#ifdef DEBUG
	DebugPrint(TRACE, DBG_PNP, "-->StartDevice()\n");
#endif

	stack = IoGetCurrentIrpStackLocation(Irp);

	PAGED_CODE();

	status = AllocateDeviceResources(FdoData, Irp);

	if (!NT_SUCCESS(status)) {
#ifdef DEBUG
		DebugPrint(ERROR, DBG_INIT, "AllocateDeviceResources failed: 0x%x\n",
								status);
#endif
		return status;
	}

	//
	// Start is an implicit power-up. Since we are done programming the
	// hardware, we should be in D0.
	//
	FdoData->DevicePowerState = PowerDeviceD0;

	powerState.DeviceState = PowerDeviceD0;
	PoSetPowerState(FdoData->Self, DevicePowerState, powerState);

	//
	// Get actual PCI Clock Frequency
	//
	status = GetPCIFrequency(FdoData);

	if (!NT_SUCCESS(status)) {
#ifdef DEBUG
		DebugPrint(ERROR, DBG_INIT, "GetPCIFrequency failed, PCIFreq %d\n",
								FdoData->PCIFreq);
#endif
		return status;
	}

	//
	// Enable the device interface. Return status
	// STATUS_OBJECT_NAME_EXISTS means we are enabling the interface
	// that was already enabled, which could happen if the device
	// is stopped and restarted for resource rebalancing.
	//
	status = IoSetDeviceInterfaceState(&FdoData->InterfaceName, TRUE);

	if (!NT_SUCCESS(status)) {
#ifdef DEBUG
		DebugPrint(ERROR, DBG_INIT, "IoSetDeviceInterfaceState failed: 0x%x\n",
								status);
#endif
		return status;
	}

	SET_NEW_PNP_STATE(FdoData, Started);

	FdoData->QueueState = AllowRequests;

	ProcessQueuedRequests(FdoData);

#ifdef DEBUG
	DebugPrint(TRACE, DBG_PNP, "<--StartDevice()\n");
#endif

	return status;
}


// ******************************************************************
// * Dispatch routine for cleanup. This routine walks the Irp queue *
// * and cancels all the pending requests for which the file object *
// * is the same one that is in the Irp.                            *
// ******************************************************************
NTSTATUS Cleanup(PDEVICE_OBJECT DeviceObject, PIRP Irp)
{
	PFDO_DATA              FdoData;
	KIRQL                  oldIrql;
	LIST_ENTRY             cleanupList;
	PLIST_ENTRY            thisEntry, nextEntry, listHead;
	PIRP                   pendingIrp;
	PIO_STACK_LOCATION     pendingIrpStack, irpStack;

#ifdef DEBUG
	DebugPrint(TRACE, DBG_CREATE_CLOSE, "-->Cleanup()\n");
#endif

	FdoData = (PFDO_DATA)DeviceObject->DeviceExtension;
	IoIncrement(FdoData);

	irpStack = IoGetCurrentIrpStackLocation(Irp);

	InitializeListHead(&cleanupList);

	//
	// We must acquire queue lock first.
	//
	KeAcquireSpinLock(&FdoData->QueueLock, &oldIrql);

	//
	// Walk through the list and remove all the IRPs
	// that belong to the input irp's fileobject.
	//

	listHead = &FdoData->NewRequestsQueue;

	for (thisEntry = listHead->Flink, nextEntry = thisEntry->Flink;
			thisEntry != listHead;
			thisEntry = nextEntry, nextEntry = thisEntry->Flink) {

		pendingIrp = CONTAINING_RECORD(thisEntry, IRP,
								Tail.Overlay.ListEntry);

		pendingIrpStack = IoGetCurrentIrpStackLocation(pendingIrp);

		if (irpStack->FileObject == pendingIrpStack->FileObject) {

			RemoveEntryList(thisEntry);

			//
			// Set the cancel routine to NULL
			//
			if (!IoSetCancelRoutine(pendingIrp, NULL)) {
				//
				// The cancel routine has run but it must be waiting to hold
				// the queue lock. It will cancel the IRP as soon as we
				// drop the lock outside this loop. We will initialize
				// the IRP's listEntry so that the cancel routine wouldn't
				// barf when it tries to remove the IRP from the queue, and
				// leave the this IRP alone.
				//
				InitializeListHead(thisEntry);

			} else {
				//
				// Cancel routine is not called and will never be
				// called. So we queue the IRP in the cleanupList
				// and cancel it after dropping the loc
				//
				InsertTailList(&cleanupList, thisEntry);
			}
		}
	}

	KeReleaseSpinLock(&FdoData->QueueLock, oldIrql);

	//
	// Walk through the cleanup list and cancel all IRPs.
	//

	while (!IsListEmpty(&cleanupList)) {

		//
		// Complete IRPs with STATUS_CANCELLED until the queue is empty.
		//

		thisEntry = RemoveHeadList(&cleanupList);
		pendingIrp = CONTAINING_RECORD(thisEntry, IRP, Tail.Overlay.ListEntry);
		pendingIrp->IoStatus.Information = 0;
		pendingIrp->IoStatus.Status = STATUS_CANCELLED;
		IoCompleteRequest(pendingIrp, IO_NO_INCREMENT);
	}

	Irp->IoStatus.Information = 0;
	Irp->IoStatus.Status = STATUS_SUCCESS;
	IoCompleteRequest(Irp, IO_NO_INCREMENT);
	IoDecrement(FdoData);

#ifdef DEBUG
	DebugPrint(TRACE, DBG_CREATE_CLOSE, "<--Cleanup()\n");
#endif

	return STATUS_SUCCESS;
}


// ***********************************************************************
// * Driver unload routine. Free all resources allocated in DriverEntry. *
// ***********************************************************************
VOID Unload(PDRIVER_OBJECT DriverObject)
{
	PAGED_CODE();

	//
	// The device object(s) should be NULL now
	// (since we unload, all the devices objects associated with this
	// driver must have already been deleted.
	//
	ASSERT(DriverObject->DeviceObject == NULL);

#ifdef DEBUG
	DebugPrint(INFO, DBG_INIT, "<-->Unload()\n");
#endif

	return;
}


// ************************************************************************
// * This function queues IRPs in the NewRequestsQueue. This queue is     *
// * used to temporarily hold the request when the device is changing     *
// * PNP state. For example:                                              *
// * 1) All the incoming requests are queued here when the device         *
// *    is stopped (IRP_MN_QUERY_STOP_DEVICE) for resource rebalancing.   *
// * 2) Requests are queued when the device is in the process of          *
// *    being query-removed (IRP_MN_QUERY_REMOVE_DEVICE).                 *
// * 3) All the incoming requests are queued when the device is in a      *
// *    low power state.                                                  *
// * -------------------------------------------------------------------- *
// * N.B. This all does NOT apply to us, it's just to satisfy WDM...      *
// ************************************************************************
NTSTATUS QueueRequest(PFDO_DATA FdoData, PIRP Irp)
{
	KIRQL               oldIrql;
	PDRIVER_CANCEL      ret;

#ifdef DEBUG
	DebugPrint(TRACE, DBG_QUEUEING, "-->QueueRequest()\n");
#endif

	//
	// Check whether we are allowed to queue requests.
	//
	ASSERT(FdoData->QueueState == HoldRequests);

	//
	// Mark the incoming IRP as pending. The incoming IRP must be marked
	// pending before setting the cancel routine so it can be completed
	// in a different thread context.
	//
	IoMarkIrpPending(Irp);

	KeAcquireSpinLock(&FdoData->QueueLock, &oldIrql);

	//
	// Set the CancelRoutine so that it can be cancelled
	// when it's waiting in the queue. Please note that
	// you must mark the irp pending before you set the
	// CancelRoutine so that the IRP can be completed in
	// a different thread context.
	//
	IoSetCancelRoutine(Irp, CancelQueued);

	//
	// Before we queue the request, let us check to see
	// if the IRP has been cancelled.
	//
	if (Irp->Cancel) {
		//
		// Yes, it is. Check whether the cancel routine
		// is already called.
		//
		ret = IoSetCancelRoutine(Irp, NULL);

		//
		// Initialize the listhead so that in case the CancelRoutine
		// is already invoked, it will not barf while trying to
		// remove the IRP from the queue after QueueLock is released
		// (below).
		//
		InitializeListHead(&Irp->Tail.Overlay.ListEntry);

		KeReleaseSpinLock(&FdoData->QueueLock, oldIrql);

		if (ret) {
			//
			// CancelRoutine hasn't been invoked and it will
			// never be as we have reset it to NULL.
			//

			Irp->IoStatus.Status = STATUS_CANCELLED;
			Irp->IoStatus.Information = 0;
			IoCompleteRequest (Irp, IO_NO_INCREMENT);
		}

	} else {
		//
		// IRP is not cancelled, so insert it in the queue.
		//
		InsertTailList(&FdoData->NewRequestsQueue,
												&Irp->Tail.Overlay.ListEntry);
		KeReleaseSpinLock(&FdoData->QueueLock, oldIrql);
	}

	IoDecrement(FdoData);

#ifdef DEBUG
	DebugPrint(TRACE, DBG_QUEUEING, "<--QueueRequest()\n");
#endif

	//
	// Since we marked the IRP pending, we must return
	// STATUS_PENDING even though the IRP has been
	// completed with STATUS_CANCELLED.
	//
	return STATUS_PENDING;
}


// ****************************************************************
// * Removes and processes the entries in the queue. Requests are *
// * either redispatched for processing or failed if the device   *
// * is in a Deleted State.                                       *
// ****************************************************************
VOID ProcessQueuedRequests(PFDO_DATA FdoData)
{

	KIRQL			oldIrql;
	PIRP			nextIrp;
	PLIST_ENTRY		listEntry;
	ULONG			nIrpsReDispatched = 0; // For debugging purposes.
	PDRIVER_CANCEL	cancelRoutine;

#ifdef DEBUG
	DebugPrint(TRACE, DBG_QUEUEING, "-->ProcessQueuedRequests()\n");
#endif

	//
	// Dequeue all the entries in the queue, reset the cancel
	// routine for each of them and process them as follows
	// 		if the device is active, send them down the stack
	// 		else complete them with STATUS_NO_SUCH_DEVICE
	//

	for(;;) {
		//
		// Acquire the queue lock before manipulating the list entries.
		//
		KeAcquireSpinLock(&FdoData->QueueLock, &oldIrql);

		if(IsListEmpty(&FdoData->NewRequestsQueue)) {
			KeReleaseSpinLock(&FdoData->QueueLock, oldIrql);
			break;
		}

		//
		// Remove a request from the queue.
		//
		listEntry = RemoveHeadList(&FdoData->NewRequestsQueue);

		nextIrp = CONTAINING_RECORD(listEntry, IRP, Tail.Overlay.ListEntry);

		cancelRoutine = IoSetCancelRoutine(nextIrp, NULL);

		//
		// Check to see if it's cancelled.
		//
		if (nextIrp->Cancel) {
			//
			// The I/O manager sets the CancelRoutine pointer in the IRP
			// to NULL before calling the routine. So check the previous
			// value to decide whether cancel routine has been
			// called or not.
			//
			if (cancelRoutine) {
				//
				// This IRP was just cancelled but the cancel routine
				// hasn't been called yet. So it's safe to cancel the IRP,
				// Let's complete the IRP after releasing the lock.
				// This is to ensure that we don't call out of the driver
				// while holding a lock.
				//
				KeReleaseSpinLock(&FdoData->QueueLock, oldIrql);
				nextIrp->IoStatus.Information = 0;
				nextIrp->IoStatus.Status = STATUS_CANCELLED;
				IoCompleteRequest(nextIrp, IO_NO_INCREMENT);
			} else {
				//
				// The cancel routine has run but it must be waiting to hold
				// the queue lock. It will cancel the IRP as soon as we
				// drop the lock. So initialize the IRPs
				// listEntry so that the cancel routine wouldn't barf
				// when it tries to remove the IRP from the queue.
				//
				InitializeListHead(listEntry);
				KeReleaseSpinLock(&FdoData->QueueLock, oldIrql);
			}
		} else {
			//
			// Release the lock before we call out of the driver
			//
			KeReleaseSpinLock(&FdoData->QueueLock, oldIrql);

			if (FdoData->QueueState == FailRequests) {
				//
				// The device was removed, we need to fail the request
				//
				nextIrp->IoStatus.Information = 0;
				nextIrp->IoStatus.Status = STATUS_NO_SUCH_DEVICE;
				IoCompleteRequest (nextIrp, IO_NO_INCREMENT);

			} else if (FdoData->QueueState == HoldRequests) {
				//
				// QueueStatus has changed again to HoldRequest, so queue
				// the current IRP back and break out of the loop.
				// Increment the I/O count to compensate for the decrement
				// done in the QueueRequest routine.
				//
				IoIncrement(FdoData);
				QueueRequest(FdoData, nextIrp);
				break;

			} else {
				//
				// Re-dispatch the IRP.
				//
				DispatchIoctl(FdoData->Self, nextIrp);
				nIrpsReDispatched++;
			}
		}
	}

#ifdef DEBUG
	DebugPrint(TRACE, DBG_QUEUEING, "<--ProcessQueuedRequests() %d\n", nIrpsReDispatched);
#endif

	return;
}


// ***************************************************************
// * Cancel routine for IRPs waiting in the NewRequestsQueue.    *
// * It will remove the IRP from the queue and will complete it. *
// * Cancel spinlock is already acquired when this routine is    *
// * called.                                                     *
// ***************************************************************
VOID CancelQueued(PDEVICE_OBJECT DeviceObject, PIRP Irp)
{
	PFDO_DATA FdoData = DeviceObject->DeviceExtension;

#ifdef DEBUG
	DebugPrint(TRACE, DBG_QUEUEING, "-->CancelQueued()\n");
#endif

	//
	// Release the cancel spinlock but stay at DISPATCH_LEVEL
	// because we are going to acquire spinlock again.
	//
	IoReleaseCancelSpinLock(DISPATCH_LEVEL);

	//
	// Acquire the local spinlock and take advantage of the fact that you
	// are running at DPC level.
	//
	KeAcquireSpinLockAtDpcLevel(&FdoData->QueueLock);

	//
	// Remove the cancelled IRP from queue and
	// release the queue lock.
	//
	RemoveEntryList(&Irp->Tail.Overlay.ListEntry);

	//
	// Release the spinlock but stay at DISPATCH_LEVEL
	//
	KeReleaseSpinLockFromDpcLevel(&FdoData->QueueLock);

	//
	// Finally lower the IRQL to the value (Irp->CancelIrql) it was
	// when the I/O manager called our CancelRoutine.
	//
	KeLowerIrql(Irp->CancelIrql);

	//
	// Complete the request with STATUS_CANCELLED.
	//
	Irp->IoStatus.Status = STATUS_CANCELLED;
	Irp->IoStatus.Information = 0;
	IoCompleteRequest(Irp, IO_NO_INCREMENT);

#ifdef DEBUG
	DebugPrint(TRACE, DBG_QUEUEING, "<--CancelQueued()\n");
#endif

	return;
}


// *************************************************************
// * Cancel the pending ioctl IRPs registered by the app or    *
// * another driver. Assumption: CancelSpinLock is already     *
// * acquired.                                                 *
// * This one is specially tailored for a particular IOCTL,    *
// * SOFT_ALLOCATE_BUFF. That one never completes, it is kept  *
// * in PENDING state until cancelled. This IOCTL is used for  *
// * mapping user app buffer into kernel virtual space.        *
// * The app opens device with FILE_FLAG_OVERLAPPED flag and   *
// * then uses SOFT_ALLOCATE_BUFF IOCLT with OVERLAPPED        *
// * structure. That IOCTL returns FALSE with error set to     *
// * ERROR_IO_PENDING. To unmap the buffer explicitely the app *
// * calls CancelIo() with file handle (or CancelIoEx() with   *
// * that OVERLAPPED structure.) It is cancelled automatically *
// *  when the App terminated.                                 *
// *************************************************************
VOID CancelRoutineForMapBufIoctlIrp(PDEVICE_OBJECT DeviceObject, PIRP Irp)
{
	PFDO_DATA			FdoData = (PFDO_DATA) DeviceObject->DeviceExtension;
	PIO_STACK_LOCATION	irpStack = NULL;

#ifdef DEBUG
	DebugPrint(TRACE, DBG_IOCTLS, "-->CancelMapBufIoctlIrp()\n");
#endif

	IoReleaseCancelSpinLock(DISPATCH_LEVEL);

	irpStack = IoGetCurrentIrpStackLocation(Irp);

	KeSynchronizeExecution(FdoData->Interrupt, HWShutdown, FdoData);

	//
	// Lower the IRQL to the value (Irp->CancelIrql) it was
	// when the I/O manager called our CancelRoutine.
	//
	KeLowerIrql(Irp->CancelIrql);

	Irp->IoStatus.Information = 0;
	Irp->IoStatus.Status = STATUS_CANCELLED;
	IoCompleteRequest(Irp, IO_NO_INCREMENT);
	IoDecrement(FdoData);

	FdoData->MapBufIrp = NULL;
	FdoData->MachSharedBlock = NULL;

#ifdef DEBUG
	DebugPrint(TRACE, DBG_IOCTLS, "<--CancelMapBufIoctlIrp\n");
#endif

	return;
}


// *************************************************
// * Same as above but for TimeData mapping IOCTL. *
// *************************************************
VOID CancelRoutineForMapTDIoctlIrp(PDEVICE_OBJECT DeviceObject, PIRP Irp)
{
	PFDO_DATA			FdoData = (PFDO_DATA) DeviceObject->DeviceExtension;
	PIO_STACK_LOCATION	irpStack = NULL;

#ifdef DEBUG
	DebugPrint(TRACE, DBG_IOCTLS, "-->CancelMapTDIoctlIrp()\n");
#endif

	IoReleaseCancelSpinLock(DISPATCH_LEVEL);

	irpStack = IoGetCurrentIrpStackLocation(Irp);

	KeSynchronizeExecution(FdoData->Interrupt, HWShutdown, FdoData);

	//
	// Lower the IRQL to the value (Irp->CancelIrql) it was
	// when the I/O manager called our CancelRoutine.
	//
	KeLowerIrql(Irp->CancelIrql);

	Irp->IoStatus.Information = 0;
	Irp->IoStatus.Status = STATUS_CANCELLED;
	IoCompleteRequest(Irp, IO_NO_INCREMENT);
	IoDecrement (FdoData);

	FdoData->MapTDIrp = NULL;
	FdoData->MachTDBlock = NULL;

#ifdef DEBUG
	DebugPrint(TRACE, DBG_IOCTLS, "<--CancelMapTDIoctlIrp()\n");
#endif

	return;
}


// *************************************
// * Return all resources driver holds *
// *************************************
NTSTATUS ReturnResources(PDEVICE_OBJECT DeviceObject)
{

	PFDO_DATA       FdoData = (PFDO_DATA) DeviceObject->DeviceExtension;
	NTSTATUS        status;
	POWER_STATE     powerState;

#ifdef DEBUG
	DebugPrint(INFO, DBG_PNP, "-->ReturnResources()\n");
#endif

	//
	// Make sure we are called to free resources when the device is stopped
	// query-removed or suprise-removed.
	//
	ASSERT(FdoData->DevicePnPState == Stopped ||
		FdoData->DevicePnPState == SurpriseRemovePending ||
		FdoData->DevicePnPState == Deleted);

	//
	// Touch the hardware to shutdown only if the device is
	// not surprise-removed.
	//
	if (FdoData->DevicePnPState != SurpriseRemovePending) {
		//
		// Remove is an implicit power-down. Since the PCI bus driver
		// writes to the power register to put the device in D3, we
		// will not worry about powering down. We just inform the system
		// about the new device power state.
		//
		if (FdoData->DevicePnPState == Deleted) {
			powerState.DeviceState = PowerDeviceD3;
			PoSetPowerState(FdoData->Self, DevicePowerState, powerState);
		}
		//
		// Reset and put the device into a known initial state.
		//
		HWShutdown(FdoData);
	}

	status = FreeDeviceResources(FdoData);

#ifdef DEBUG
	DebugPrint(INFO, DBG_PNP, "<--ReturnResources()\n");
#endif

	return status;
}


// ********************************************
// * Sends the Irp down the stack and doesn't *
// * care what happens to the IRP after that. *
// ********************************************
NTSTATUS ForwardAndForget(PFDO_DATA FdoData, PIRP Irp)
{
	NTSTATUS		status;

	IoSkipCurrentIrpStackLocation(Irp);
	status = IoCallDriver(FdoData->NextLowerDriver, Irp);
	IoDecrement(FdoData);

	return status;
}


// ***************************************************
// * Send IRP down stack and wait for it to complete *
// ***************************************************
NTSTATUS SendIrpSynchronously(PDEVICE_OBJECT DeviceObject, PIRP Irp)
{
	KEVENT				event;
	NTSTATUS			status;

	PAGED_CODE();

	KeInitializeEvent(&event, NotificationEvent, FALSE);

	IoCopyCurrentIrpStackLocationToNext(Irp);

	IoSetCompletionRoutine(Irp, DispatchPnpComplete, &event,
							TRUE, TRUE, TRUE);

	status = IoCallDriver(DeviceObject, Irp);

	if (status == STATUS_PENDING) {
		status = KeWaitForSingleObject(&event, Executive, KernelMode, FALSE, NULL);
		ASSERT(NT_SUCCESS(status));
		status = Irp->IoStatus.Status;
	}

	return status;
}


// *****************************************************************
// * This routine is used to queue workitems so that the callback  *
// * functions can be executed at PASSIVE_LEVEL in the context of  *
// * a system thread. It is used in power.c.                       *
// *****************************************************************
NTSTATUS QueuePassiveLevelCallback(PFDO_DATA FdoData,
		PIO_WORKITEM_ROUTINE CallbackFunction, PVOID Context1, PVOID Context2)
{
	PIO_WORKITEM		item = NULL;
	NTSTATUS			status = STATUS_SUCCESS;
	PWORK_ITEM_CONTEXT	context;

	context = ExAllocatePoolWithTag(NonPagedPool, sizeof(WORKER_ITEM_CONTEXT),
										CPDDRV_POOL_TAG);

	if (!context) return STATUS_INSUFFICIENT_RESOURCES;

	//
	// Allocate a workitem and queue it for passive level processing.
	//
	item = IoAllocateWorkItem(FdoData->Self);

	if (item) {
		context->WorkItem = item;
		context->Argument1 = Context1;
		context->Argument2 = Context2;

		IoQueueWorkItem(item, CallbackFunction, DelayedWorkQueue, context);
	} else {
		status = STATUS_INSUFFICIENT_RESOURCES;
	}

	return status;
}


// *******************************
// * Increments driver I/O count *
// *******************************
LONG IoIncrement(PFDO_DATA FdoData)
{
	LONG	result;

	ASSERT(FdoData->OutstandingIO >= 1);

	result = InterlockedIncrement(&FdoData->OutstandingIO);

#ifdef DEBUG
	DebugPrint(TRACE, DBG_LOCKS, "IoIncrement --> %d\n", result);
#endif

	return result;
}


// *********************************************************************
// * This routine decrements outstanding I/O count as it completes a   *
// * request. It also performs two special actions depending on the    *
// * resultant outstanding count. If the count equals 1 that indicates *
// * the device is stopping. If the count equals 0 that indicates the  *
// * device is being removed. Returns outstanding I/O count.           *
// *********************************************************************
LONG IoDecrement(PFDO_DATA FdoData)
{
	LONG	result;

	ASSERT(FdoData->OutstandingIO >= 1);

	result = InterlockedDecrement(&FdoData->OutstandingIO);

#ifdef DEBUG
	DebugPrint(TRACE, DBG_LOCKS, "IoDecrement --> %d\n", result);
#endif

	if (result == 1) {
		//
		// The count is 2-biased, so it can be 1 only if an
		// extra decrement is done when the device needs to be stopped
		// for resource rebalancing or powered down.
		// Note that when this happens, any new requests we want to be
		// processed are already held in the queue instead of being
		// passed away, so that we can't "miss" a request that
		// will appear between the decrement and the moment when
		// the value is actually used.
		//
		KeSetEvent (&FdoData->StopEvent, IO_NO_INCREMENT, FALSE);
	}

	if (result == 0) {
		//
		// The count is 2-biased, so it can be zero only if two
		// extra decrement is done when a remove Irp is received
		//

		ASSERT(FdoData->DevicePnPState == Deleted);
		//
		// Set the remove event, so the device object can be deleted.
		//
		KeSetEvent (&FdoData->RemoveEvent, IO_NO_INCREMENT, FALSE);
	}

	return result;
}


// ***********************************************************
// * This routine waits for all outstanding requests, except *
// * the count specified in the OnHoldCount, to finish.      *
// ***********************************************************
VOID ReleaseAndWait(PFDO_DATA FdoData, ULONG OnHoldCount, WAIT_REASON Reason)
{
	NTSTATUS	status;
	ULONG		chargeRemaining;

	ASSERT(OnHoldCount > 0);

	if (Reason == DEV_STOP) {
		//
		// If the wait reason is STOP, we will do one extra decrement so
		// that the OustandingIo count can come down to 1 when all the
		// other requests in progress are finished. At that point, the Stop
		// event will be signalled to indicate there are no I/Os currently
		// active.
		//
		chargeRemaining =  OnHoldCount + 1;

		while (chargeRemaining--) IoDecrement(FdoData);

#ifdef DEBUG
		DebugPrint(INFO, DBG_PNP,
			"Waiting for pending requests to complete (Stop)...\n");
#endif

		status = KeWaitForSingleObject(&FdoData->StopEvent, Executive,
										KernelMode, FALSE, NULL);

		ASSERT(NT_SUCCESS(status));

		ASSERT(FdoData->OutstandingIO == 1);
		//
		// Since STOP is a transient state, we should increment the
		// OutstandingIo count to account for requests currently held
		// by the caller and  also rebase the count to 2 by doing
		// an extra increment.
		//
		chargeRemaining =  OnHoldCount + 1;

		while(chargeRemaining--) IoIncrement(FdoData);

	} else if (Reason == DEV_REMOVE) {
		//
		// If the wait reason is REMOVE, we will do two extra decrements
		// so that the OustandingIo count can come down to 0 when all the
		// other requests in progress are finished. At that point, the
		// Remove event will be signalled to indicate there are no I/Os
		// currently active. Remove is a destructive state. After this
		// no I/O requests will be processed on this device and the
		// device object will be deleted.
		//
		chargeRemaining =  OnHoldCount + 2;

		while(chargeRemaining--) IoDecrement(FdoData);

#ifdef DEBUG
		DebugPrint(INFO, DBG_PNP,
			"Waiting for pending requests to complete (Remove)...\n");
#endif

		status = KeWaitForSingleObject(&FdoData->RemoveEvent, Executive,
										KernelMode, FALSE, NULL);

		ASSERT(NT_SUCCESS(status));
		ASSERT(FdoData->OutstandingIO == 0);

	} else {
		ASSERTMSG("Unknown reason", FALSE);
	}
	return;
}


// ***********************************************************************
// * This is where all IOCTL IRPs get sent. No Read/Write for our driver *
// ***********************************************************************
NTSTATUS DispatchIoctl (PDEVICE_OBJECT DeviceObject, PIRP Irp)
{
	PIO_STACK_LOCATION      	irpStack;
	NTSTATUS                	status = STATUS_SUCCESS;
	PFDO_DATA               	FdoData;
	ULONG                   	FunctionCode;
	ULONG                   	bytesReturned;
	UINT32						tmp;
	INT							i;
	PVOID						appBuf;

#ifdef DEBUG
	DebugPrint(TRACE, DBG_IOCTLS, "-->DispatchIoctl() %p\n", Irp);
#endif

	FdoData = (PFDO_DATA)DeviceObject->DeviceExtension;
	irpStack = IoGetCurrentIrpStackLocation(Irp);

	FunctionCode = irpStack->Parameters.DeviceIoControl.IoControlCode;

	IoIncrement(FdoData);

	if (FdoData->DevicePnPState == Deleted) {
		//
		// Since the device is removed, we will fail the request.
		//
		status = STATUS_NO_SUCH_DEVICE;
		goto EndIOCTL;
	}

	//
	// We only support IOCTLs so all others are failed.
	//
	if (irpStack->MajorFunction != IRP_MJ_DEVICE_CONTROL) {
		ASSERTMSG(FALSE, "DispatchIoctl() invalid IRP");
		status = STATUS_NOT_SUPPORTED;
		goto EndIOCTL;
	}

	//
	// This should not happen because we don't do Power Management but...
	//
	if (FdoData->QueueState == HoldRequests)
		return QueueRequest(FdoData, Irp);

	//
	// Overwrite it if different length. Most IOCTLs return just one INT.
	// Also status is set to STATUS_SUCCESS at the top so do not make
	// any assignments to it unless non-success status is required.
	//
	bytesReturned = sizeof(INT);

	appBuf = Irp->AssociatedIrp.SystemBuffer;

	//
	// OK, this is IOCTL so go process it.
	//
	switch (FunctionCode) {

		case MACH1_PREPTIMER:
			// **************
			// Do nothing, always report success. It was probably used for
			// something in Mach3 but no longer required so it is a dummy now.
			// **************
#ifdef DEBUG
			DebugPrint(TRACE, DBG_IOCTLS, "IOCTL: MACH1_PREPTIMER\n");
#endif
			ASSERT((FunctionCode & 0x3) == METHOD_BUFFERED);
			((PINT)appBuf)[0] = MACH1_SUCCESS;
			break;

		case MACH1_GETCOUNT:
			// !!!!!!!!!!!!!
			// KSI: Don't know what it is and what is it used for. So far.
			// !!!!!!!!!!!!!
#ifdef DEBUG
			DebugPrint(TRACE, DBG_IOCTLS, "IOCTL: MACH1_GETCOUNT\n");
#endif
			status = STATUS_NOT_SUPPORTED;
			break;

		case MACH1_GETMAXTIME:
			// @@@@@@@@@@@@@@@
			// KSI: Not in Art's driver, caught by default case there.
			// @@@@@@@@@@@@@@@
#ifdef DEBUG
			DebugPrint(TRACE, DBG_IOCTLS, "IOCTL: MACH1_GETMAXTIME\n");
#endif
			status = STATUS_NOT_SUPPORTED;
			break;
		case MACH1_GETMINTIME:
			// @@@@@@@@@@@@@@@
			// KSI: Not in Art's driver, caught by default case there.
			// @@@@@@@@@@@@@@@
#ifdef DEBUG
			DebugPrint(TRACE, DBG_IOCTLS, "IOCTL: MACH1_GETMINTIME\n");
#endif
			status = STATUS_NOT_SUPPORTED;
			break;

		//
		// This one starts the show after buffers were mapped. It sets up
		// I/O signals mapping according to Mach3 configuration and starts
		// the driver in DrvEStopped state @ 25KHz kernel frequency.
		//
		case MACH1_STARTTIMER:
#ifdef DEBUG
			DebugPrint(TRACE, DBG_IOCTLS, "IOCTL: MACH1_STARTTIMER\n");
#endif
			ASSERT((FunctionCode & 0x3) == METHOD_BUFFERED);

			((PINT)appBuf)[0] = MACH1_SUCCESS;

			if (FdoData->Running) break;	// Already running

			if (FdoData->MachSharedBlock) {
				//
				// OK, start timer @ 25KHz if Mach3 shared block is mapped.
				//
				if (!KeSynchronizeExecution(FdoData->Interrupt,
											MachDoIoctlStartTimer, FdoData))
					((PINT)appBuf)[0] = PULSE_ERROR_NOTIMER;

			} else {
				//
				// No buffer, send a failure code to the App (TimerNotStarted).
				//
				((PINT)appBuf)[0] = PULSE_ERROR_NOTIMER;
			}

			break;

		case SOFT_GET_STATUS:
			// @@@@@@@@@@@@@@@
			// KSI: Not in Art's driver, caught by default case there.
			// @@@@@@@@@@@@@@@
#ifdef DEBUG
			DebugPrint(TRACE, DBG_IOCTLS, "IOCTL: SOFT_GET_STATUS\n");
#endif
			status = STATUS_NOT_SUPPORTED;
			break;
		case SOFT_CONTROL_OUT:
			// @@@@@@@@@@@@@@@
			// KSI: Not in Art's driver, caught by default case there.
			// @@@@@@@@@@@@@@@
#ifdef DEBUG
			DebugPrint(TRACE, DBG_IOCTLS, "IOCTL: SOFT_CONTROL_OUT\n");
#endif
			status = STATUS_NOT_SUPPORTED;
			break;
		case SOFT_SET_FREQUENCY:
			// @@@@@@@@@@@@@@@
			// KSI: Not in Art's driver, caught by default case there.
			// @@@@@@@@@@@@@@@
#ifdef DEBUG
			DebugPrint(TRACE, DBG_IOCTLS, "IOCTL: SOFT_SET_FREQUENCY\n");
#endif
			status = STATUS_NOT_SUPPORTED;
			break;
		case SOFT_ACTIVATE:
			// @@@@@@@@@@@@@@@
			// KSI: Not in Art's driver, caught by default case there.
			// @@@@@@@@@@@@@@@
#ifdef DEBUG
			DebugPrint(TRACE, DBG_IOCTLS, "IOCTL: SOFT_ACTIVATE\n");
#endif
			status = STATUS_NOT_SUPPORTED;
			break;
		case SOFT_GET_TIMES:
			// @@@@@@@@@@@@@@@
			// KSI: Not in Art's driver, caught by default case there.
			// @@@@@@@@@@@@@@@
#ifdef DEBUG
			DebugPrint(TRACE, DBG_IOCTLS, "IOCTL: SOFT_GET_TIMES\n");
#endif
			status = STATUS_NOT_SUPPORTED;
			break;
		case SOFT_GET_DATA:
			// @@@@@@@@@@@@@@@
			// KSI: Not in Art's driver, caught by default case there.
			// @@@@@@@@@@@@@@@
#ifdef DEBUG
			DebugPrint(TRACE, DBG_IOCTLS, "IOCTL: SOFT_GET_DATA\n");
#endif
			status = STATUS_NOT_SUPPORTED;
			break;
		case SOFT_ADD_TO_BUFFER:
			// @@@@@@@@@@@@@@@
			// KSI: Not in Art's driver, caught by default case there.
			// @@@@@@@@@@@@@@@
#ifdef DEBUG
			DebugPrint(TRACE, DBG_IOCTLS, "IOCTL: SOFT_ADD_TO_BUFFER\n");
#endif
			status = STATUS_NOT_SUPPORTED;
			break;
		case SOFT_GET_LOADINDEX:
			// @@@@@@@@@@@@@@@
			// KSI: Not in Art's driver, caught by default case there.
			// @@@@@@@@@@@@@@@
#ifdef DEBUG
			DebugPrint(TRACE, DBG_IOCTLS, "IOCTL: SOFT_GET_LOADINDEX\n");
#endif
			status = STATUS_NOT_SUPPORTED;
			break;
		case SOFT_GET_CONTROLINDEX:
			// @@@@@@@@@@@@@@@
			// KSI: Not in Art's driver, caught by default case there.
			// @@@@@@@@@@@@@@@
#ifdef DEBUG
			DebugPrint(TRACE, DBG_IOCTLS, "IOCTL: SOFT_CONTROLINDEX\n");
#endif
			status = STATUS_NOT_SUPPORTED;
			break;
		case SOFT_SET_PORT:
			// @@@@@@@@@@@@@@@
			// KSI: Not in Art's driver, caught by default case there.
			// @@@@@@@@@@@@@@@
#ifdef DEBUG
			DebugPrint(TRACE, DBG_IOCTLS, "IOCTL: SOFT_SET_PORT\n");
#endif
			status = STATUS_NOT_SUPPORTED;
			break;
		case SOFT_LOAD_MASKS:
			// @@@@@@@@@@@@@@@
			// KSI: Not in Art's driver, caught by default case there.
			// @@@@@@@@@@@@@@@
#ifdef DEBUG
			DebugPrint(TRACE, DBG_IOCTLS, "IOCTL: SOFT_LOAD_MASKS\n");
#endif
			status = STATUS_NOT_SUPPORTED;
			break;
		case SOFT_GET_ACTIVE:
			// @@@@@@@@@@@@@@@
			// KSI: Not in Art's driver, caught by default case there.
			// @@@@@@@@@@@@@@@
#ifdef DEBUG
			DebugPrint(TRACE, DBG_IOCTLS, "IOCTL: SOFT_GET_ACTIVE\n");
#endif
			status = STATUS_NOT_SUPPORTED;
			break;
		case SOFT_GETINDEXS:
			// @@@@@@@@@@@@@@@
			// KSI: Not in Art's driver, caught by default case there.
			// @@@@@@@@@@@@@@@
#ifdef DEBUG
			DebugPrint(TRACE, DBG_IOCTLS, "IOCTL: SOFT_GETINDEXS\n");
#endif
			status = STATUS_NOT_SUPPORTED;
			break;
		case SOFT_SETREVERSALS:
			// @@@@@@@@@@@@@@@
			// KSI: Not in Art's driver, caught by default case there.
			// @@@@@@@@@@@@@@@
#ifdef DEBUG
			DebugPrint(TRACE, DBG_IOCTLS, "IOCTL: SOFT_SETREVERSALS\n");
#endif
			status = STATUS_NOT_SUPPORTED;
			break;
		case SOFT_CONTROL_IN:
			// @@@@@@@@@@@@@@@
			// KSI: Not in Art's driver, caught by default case there.
			// @@@@@@@@@@@@@@@
#ifdef DEBUG
			DebugPrint(TRACE, DBG_IOCTLS, "IOCTL: SOFT_CONTROL_IN\n");
#endif
			status = STATUS_NOT_SUPPORTED;
			break;
		case SOFT_CONTROL_JOGON:
			// @@@@@@@@@@@@@@@
			// KSI: Not in Art's driver, caught by default case there.
			// @@@@@@@@@@@@@@@
#ifdef DEBUG
			DebugPrint(TRACE, DBG_IOCTLS, "IOCTL: SOFT_CONTROL_JOGON\n");
#endif
			status = STATUS_NOT_SUPPORTED;
			break;
		case SOFT_CONTROL_JOGOFF:
			// @@@@@@@@@@@@@@@
			// KSI: Not in Art's driver, caught by default case there.
			// @@@@@@@@@@@@@@@
#ifdef DEBUG
			DebugPrint(TRACE, DBG_IOCTLS, "IOCTL: SOFT_CONTROL_JOGOFF\n");
#endif
			status = STATUS_NOT_SUPPORTED;
			break;
		case SOFT_SETNEGS:
			// @@@@@@@@@@@@@@@
			// KSI: Not in Art's driver, caught by default case there.
			// @@@@@@@@@@@@@@@
#ifdef DEBUG
			DebugPrint(TRACE, DBG_IOCTLS, "IOCTL: SOFT_SETNEGS\n");
#endif
			status = STATUS_NOT_SUPPORTED;
			break;
		case SOFT_GETIDLE:
			// @@@@@@@@@@@@@@@
			// KSI: Not in Art's driver, caught by default case there.
			// @@@@@@@@@@@@@@@
#ifdef DEBUG
			DebugPrint(TRACE, DBG_IOCTLS, "IOCTL: SOFT_GETIDLE\n");
#endif
			status = STATUS_NOT_SUPPORTED;
			break;
		case SOFT_EMERGENCY:
			// @@@@@@@@@@@@@@@
			// KSI: Not in Art's driver, caught by default case there.
			// @@@@@@@@@@@@@@@
#ifdef DEBUG
			DebugPrint(TRACE, DBG_IOCTLS, "IOCTL: SOFT_EMERGENCY\n");
#endif
			status = STATUS_NOT_SUPPORTED;
			break;
		case SOFT_OVERRIDE:
			// @@@@@@@@@@@@@@@
			// KSI: Not in Art's driver, caught by default case there.
			// @@@@@@@@@@@@@@@
#ifdef DEBUG
			DebugPrint(TRACE, DBG_IOCTLS, "IOCTL: SOFT_OVERRIDE\n");
#endif
			status = STATUS_NOT_SUPPORTED;
			break;
		case SOFT_GOHOME:
			// @@@@@@@@@@@@@@@
			// KSI: Not in Art's driver, caught by default case there.
			// @@@@@@@@@@@@@@@
#ifdef DEBUG
			DebugPrint(TRACE, DBG_IOCTLS, "IOCTL: SOFT_GOHOME\n");
#endif
			status = STATUS_NOT_SUPPORTED;
			break;
		case SOFT_PAUSE:
			// @@@@@@@@@@@@@@@
			// KSI: Not in Art's driver, caught by default case there.
			// @@@@@@@@@@@@@@@
#ifdef DEBUG
			DebugPrint(TRACE, DBG_IOCTLS, "IOCTL: SOFT_PAUSE\n");
#endif
			status = STATUS_NOT_SUPPORTED;
			break;
		case SOFT_HALT:
			// @@@@@@@@@@@@@@@
			// KSI: Not in Art's driver, caught by default case there.
			// @@@@@@@@@@@@@@@
#ifdef DEBUG
			DebugPrint(TRACE, DBG_IOCTLS, "IOCTL: SOFT_HALT\n");
#endif
			status = STATUS_NOT_SUPPORTED;
			break;
		case SOFT_LASER:
			// @@@@@@@@@@@@@@@
			// KSI: Not in Art's driver, caught by default case there.
			// @@@@@@@@@@@@@@@
#ifdef DEBUG
			DebugPrint(TRACE, DBG_IOCTLS, "IOCTL: SOFT_LASER\n");
#endif
			status = STATUS_NOT_SUPPORTED;
			break;
		case SOFT_SPACE:
			// @@@@@@@@@@@@@@@
			// KSI: Not in Art's driver, caught by default case there.
			// @@@@@@@@@@@@@@@
#ifdef DEBUG
			DebugPrint(TRACE, DBG_IOCTLS, "IOCTL: SOFT_SPACE\n");
#endif
			status = STATUS_NOT_SUPPORTED;
			break;
		case SOFT_GATE:
			// @@@@@@@@@@@@@@@
			// KSI: Not in Art's driver, caught by default case there.
			// @@@@@@@@@@@@@@@
#ifdef DEBUG
			DebugPrint(TRACE, DBG_IOCTLS, "IOCTL: SOFT_GATE\n");
#endif
			status = STATUS_NOT_SUPPORTED;
			break;

		//
		// This one maps the Mach3 TrajBuffer to kernel memory space
		// so it can be used by the driver. That buffer is locked on
		// the IOCTL request because it is METHOD_OUT_DIRECT so it is
		// safe to map and use it. However this IOCTL must NOT complete
		// the IRP because buffer can be unlocked on completion. That
		// is why the IRP itself is saved in FdoData structure for
		// later completion on either cancel routine on application
		// termination or by the MACH1_RELEASEBUFFERS IOCTL.
		//
		case SOFT_ALLOCATE_BUFF:

#ifdef DEBUG
			DebugPrint(TRACE, DBG_IOCTLS, "IOCTL: SOFT_ALLOCATE_BUFF\n");
#endif

			//
			// It MUST be METHOD_OUT_DIRECT !
			//
			ASSERT((FunctionCode & 0x3) == METHOD_OUT_DIRECT);

			if (FdoData->MachSharedBlock) break;	// Already mapped

			//
			// Map User Buffer into Kernel Virtual Space. This is
			// a pointer to MachSharedBlock structure from Mach3 app.
			//
			FdoData->MachSharedBlock =
				(PMACH_STRUCT)MmGetSystemAddressForMdlSafe(
							Irp->MdlAddress, NormalPagePriority);

			if (!FdoData->MachSharedBlock) {
				status = STATUS_INSUFFICIENT_RESOURCES;
				break;
			}

			status = STATUS_PENDING;
			IoMarkIrpPending(Irp);
			IoSetCancelRoutine(Irp, CancelRoutineForMapBufIoctlIrp);
			FdoData->MapBufIrp = Irp;

#ifdef DEBUG
			DebugPrint(TRACE, DBG_IOCTLS, "MachSharedBlock: 0x%p\n",
						FdoData->MachSharedBlock);
#endif

			break;

		//
		// Same as above but for TimeData structure that is rarely used
		// but we do it to keep compatibility.
		//
		case SOFT_SAMPLE_BUFF:

#ifdef DEBUG
			DebugPrint(TRACE, DBG_IOCTLS, "IOCTL: SOFT_SAMPLE_BUFF\n");
#endif
			//
			// It MUST be METHOD_OUT_DIRECT !
			//
			ASSERT((FunctionCode & 0x3) == METHOD_OUT_DIRECT);

			if (FdoData->MachTDBlock) break;	// Already mapped

			//
			// Map TimeData struct into Kernel Virtual Space. This is
			// a pointer to TimeData structure from Mach3 app.
			//
			FdoData->MachTDBlock =
				(PMACH_TD_STRUCT)MmGetSystemAddressForMdlSafe(
								Irp->MdlAddress, NormalPagePriority);

			if (!FdoData->MachTDBlock) {
				status = STATUS_INSUFFICIENT_RESOURCES;
				break;
			}

			status = STATUS_PENDING;
			IoMarkIrpPending(Irp);
			IoSetCancelRoutine(Irp, CancelRoutineForMapTDIoctlIrp);
			FdoData->MapTDIrp = Irp;

#ifdef DEBUG
			DebugPrint(TRACE, DBG_IOCTLS, "MachTDBlock: 0x%p\n",
						FdoData->MachTDBlock);
#endif

			break;

		case SOFT_SAMPLE_ON:
			// !!!!!!!!!!!!!
			// KSI: This is NOT implemented so far.
			// !!!!!!!!!!!!!
#ifdef DEBUG
			DebugPrint(TRACE, DBG_IOCTLS, "IOCTL: SOFT_SAMPLE_ON\n");
#endif
			status = STATUS_NOT_SUPPORTED;
			break;

		case MACH1_STOPTIMER:
#ifdef DEBUG
			DebugPrint(TRACE, DBG_IOCTLS, "IOCTL: MACH1_STOPTIMER\n");
#endif
		case MACH1_RELEASEBUFFERS:
			// **************
			// This actually stops all activity, unmaps Mach3 shared memory
			// that was mapped to kernel space, completes that suspended
			// pending IRP for it, sets all the outputs to inactive state
			// and performs other required steps to orderly stop the driver.
			// **************
#ifdef DEBUG
			DebugPrint(TRACE, DBG_IOCTLS, "IOCTL: MACH1_RELEASEBUFFERS\n");
#endif
			ASSERT((FunctionCode & 0x3) == METHOD_BUFFERED);

			//
			// Shutdown the board. HWShutdown() takes care of HW shutdown
			// and sets driver state to NOT running/started/ready. It uses
			// IRQL spinlock so ISR won't be able to mess with it.
			//
			KeSynchronizeExecution(FdoData->Interrupt, HWShutdown, FdoData);

			if (FdoData->MapBufIrp) {
				if (IoSetCancelRoutine(FdoData->MapBufIrp, NULL)) {
					FdoData->MapBufIrp->IoStatus.Information = bytesReturned;
					FdoData->MapBufIrp->IoStatus.Status = STATUS_SUCCESS;
					IoCompleteRequest(FdoData->MapBufIrp, IO_NO_INCREMENT);
					IoDecrement(FdoData);
				}
				FdoData->MapBufIrp = NULL;
			}

			if (FdoData->MapTDIrp) {
				if (IoSetCancelRoutine(FdoData->MapTDIrp, NULL)) {
					FdoData->MapTDIrp->IoStatus.Information = bytesReturned;
					FdoData->MapTDIrp->IoStatus.Status = STATUS_SUCCESS;
					IoCompleteRequest(FdoData->MapTDIrp, IO_NO_INCREMENT);
					IoDecrement(FdoData);
				}
				FdoData->MapTDIrp = NULL;
			}

			// Destroy mapping
			FdoData->MachSharedBlock = NULL;
			FdoData->MachTDBlock = NULL;

			((PINT)appBuf)[0] = MACH1_SUCCESS;

			break;

		case MACH1_DRIVERVERSION:
			// **************
			// This one sets timer frequency if invoked with a parameter
			// between 94 and 100 or returns the driver version otherwise.
			// **************
#ifdef DEBUG
			DebugPrint(TRACE, DBG_IOCTLS, "IOCTL: MACH1_DRIVERVERSION\n");
#endif
			ASSERT((FunctionCode & 0x3) == METHOD_BUFFERED);

			i = ((PINT)appBuf)[0];

			if (i <= 100 && i >= 94) {				// Set Timer Frequency

				((PINT)appBuf)[0] = 1;

				switch (i) {

					case 94:
						// Set board timebase to to 25 KHz
						FdoData->TmrFreq = 25;
						break;

					case 95:
						// Set board timebase to to ~60 KHz
						FdoData->TmrFreq = 60;
						break;

					case 96:
						// Set board timebase to to ~65 KHz
						FdoData->TmrFreq = 65;
						break;

					case 97:
						// Set board timebase to to ~75 KHz
						FdoData->TmrFreq = 75;
						break;

					case 98:
						// Set board timebase to to 100 KHz
						FdoData->TmrFreq = 100;
						break;

					case 99:
						// Set board timebase to to ~35 KHz
						FdoData->TmrFreq = 35;
						break;

					case 100:
						// Set board timebase to to ~45 KHz
						FdoData->TmrFreq = 45;
						break;
				}

				KeSynchronizeExecution(FdoData->Interrupt, MachSetTimerFreq,
													FdoData);

			} else {								// Return Driver Version

				((PINT)appBuf)[0] = DRIVERVERSION;
			}

			status = STATUS_SUCCESS;
			bytesReturned = sizeof(INT);
			break;

		case MACH1_WAITFORTIMER:
			// !!!!!!!!!!!!!
			// KSI: This is currently NOT implemented.
			// !!!!!!!!!!!!!
#ifdef DEBUG
			DebugPrint(TRACE, DBG_IOCTLS, "IOCTL: MACH1_WAITFORTIMER\n");
#endif
			status = STATUS_NOT_SUPPORTED;
			break;

		 default:
			ASSERTMSG(FALSE, "Invalid IOCTL request\n");
			status = STATUS_NOT_SUPPORTED;
			break;
	}

EndIOCTL:

	if (status != STATUS_PENDING) {
		Irp->IoStatus.Information = bytesReturned;
		Irp->IoStatus.Status = status;
		IoCompleteRequest(Irp, IO_NO_INCREMENT);
		IoDecrement(FdoData);
	}

#ifdef DEBUG
	DebugPrint(TRACE, DBG_IOCTLS, "<--DispatchIoctl() %p\n", Irp);
#endif

	return status;
}
