/*
 * Tabs are 4 symbols wide due to stupid Windows function names...
 * ---------------------------------------------------------------
 *
 * Copyright (c) 2010 KSI Labs, LLC. All rights reserved.
 * Written by Sergey Kubushyn <ksi@koi8.net>, <ksi@ksilabs.com>.
 *
 * File: sigfuncs.c
 *
 * This is Mach3-specific routines for  WDM driver for KSI Labs, LLC
 * CENTIPEDE PCI Board. Set of optimized assembly accessors for reading and
 * writing to I/O port pins. They are all "naked" functions that saves at
 * least prolog/epilog overhead. That is substantial because there is one
 * such function per signal pin and it gets called every Centipede timer tic
 * that is every 200 uS or even less. There might be something like 60 such
 * calls or so per tick and each prolog/epilog pair has at least 4 CPU
 * instructions.
 *
 * Those functions are called with direct assembly call instruction through
 * a call table that is populated upon driver startup according to Mach3
 * configuration (pins/signals mapping.)
 *
 * Please read documentation or go to http://www.ksilabs.com for further info.
 */

/*
 * 						LICENSE and Terms of Use.
 *
 * First of all, this is NOT GPL or Free Software. It is open so you have
 * all the source code for the Engine. There is also VHDL source code for
 * CENTIPEDE board CPLD that is available from KSI Labs, LLC website,
 * http://www.ksilabs.com.
 *
 * As for this particular code, Windows Mach3 Engine, you are free to use it,
 * make any modifications to customize it for your own needs, fix inevitable
 * bugs, port it to other CNC software and do anything else as long as your
 * derivative work that is based on this code or uses parts or pieces thereof
 * is made to be used with any KSI Labs, LLC products.
 *
 * We do NOT grant you a permission to use any parts of this software in any
 * products that are designed for products not manufactured or distributed
 * by KSI Labs, LLC. We might give such a permission on case-by-case basis
 * but it is not legal to use this software or bits and pieces thereof with
 * products not coming from KSI Labs, LLC without our explicit written
 * permission.
 *
 * The original copyright notices must be included in all derivative works.
 * You do NOT have to ask our permission to make any modifications to this
 * code and you are not under obligation to give us any of your work back
 * as long as you are using it for your own needs.
 *
 * However we do NOT give our permission to distribute your derivative work
 * unless all original copyrights are retained, all the conditions of this
 * license are met (use only with KSI Labs, LLC products) and full source
 * code for such derivative work is available for free to everybody without
 * any further conditions including but not limited to signing any kinds of
 * NDA. We explicitely deny permission to distribute any derivative work
 * based on this code with closed source binary-only fragments or include
 * bits or pieces of this code in such products.
 *
 * CNC software like Mach3 and likes is excluded from these requirements as
 * long as it does not include parts or pieces of this software into the
 * close source proprietary application proper. It can use shared memory
 * buffer, IOCTL calls, etc. to communicate with this Engine or the one based
 * on it and keep all the software making those calls or working on that
 * buffer close source and proprietary as long as those calls/structures are
 * documented and that documentation is freely available to everybody without
 * any further conditions including but no limited to signing any kind of NDA.
 */

#include "precomp.h"

/*
 * All output data for 2 parallel ports are packed into 1 32-bit DWORD.
 *
 *	Bits 0..7		Port1 Data register, Base address
 *	Bits 8..15		Port1 Control register, Base + 2
 *	Bits 16..23		Port2 Data register, Base
 *	Bits 24..31		Port2 Control register, Base + 2
 *
 * All data are gathered from Mach3 OutputInfo structures and placed in that
 * DWORD first that greatly speeds up the process -- memory access is much
 * faster than port I/O.
 *
 * When populated that DWORD is output to the actual ports in 4 steps
 *
 *	Out lower byte to Port1 Data, shift right 8 positions
 *	Out lower byte to Port1 Control, shift right 8 positions
 *	Out lower byte to Port2 Data, shift right 8 positions
 *	Out lower byte to Port2 Control, shift right 8 positions
 */
#define OUT_PP1_P1	0x00000100
#define OUT_PP1_P2	0x00000001
#define OUT_PP1_P3	0x00000002
#define OUT_PP1_P4	0x00000004
#define OUT_PP1_P5	0x00000008
#define OUT_PP1_P6	0x00000010
#define OUT_PP1_P7	0x00000020
#define OUT_PP1_P8	0x00000040
#define OUT_PP1_P9	0x00000080
#define OUT_PP1_P14	0x00000200
#define OUT_PP1_P16	0x00000400
#define OUT_PP1_P17	0x00000800
#define OUT_PP2_P1	0x01000000
#define OUT_PP2_P2	0x00010000
#define OUT_PP2_P3	0x00020000
#define OUT_PP2_P4	0x00040000
#define OUT_PP2_P5	0x00080000
#define OUT_PP2_P6	0x00100000
#define OUT_PP2_P7	0x00200000
#define OUT_PP2_P8	0x00400000
#define OUT_PP2_P9	0x00800000
#define OUT_PP2_P14	0x02000000
#define OUT_PP2_P16	0x04000000
#define OUT_PP2_P17	0x08000000

// Bulk of Port Output Signal Handlers
static __declspec(naked) VOID OutSigPort1Pin1(VOID)
{
	__asm {
			xor		al, ah							; Negated ^ Activated
			jz		skip_bit						; Skip Sig Bit, it is already 0
			or		dword ptr [edi], OUT_PP1_P1		; Set signal bit
		skip_bit:
			ret
	}
}

static __declspec(naked) VOID OutSigPort1Pin2(VOID)
{
	__asm {
			xor		al, ah							; Negated ^ Activated
			jz		skip_bit						; Skip Sig Bit, it is already 0
			or		dword ptr [edi], OUT_PP1_P2		; Set signal bit
		skip_bit:
			ret
	}
}

static __declspec(naked) VOID OutSigPort1Pin3(VOID)
{
	__asm {
			xor		al, ah							; Negated ^ Activated
			jz		skip_bit						; Skip Sig Bit, it is already 0
			or		dword ptr [edi], OUT_PP1_P3		; Set signal bit
		skip_bit:
			ret
	}
}

static __declspec(naked) VOID OutSigPort1Pin4(VOID)
{
	__asm {
			xor		al, ah							; Negated ^ Activated
			jz		skip_bit						; Skip Sig Bit, it is already 0
			or		dword ptr [edi], OUT_PP1_P4		; Set signal bit
		skip_bit:
			ret
	}
}

static __declspec(naked) VOID OutSigPort1Pin5(VOID)
{
	__asm {
			xor		al, ah							; Negated ^ Activated
			jz		skip_bit						; Skip Sig Bit, it is already 0
			or		dword ptr [edi], OUT_PP1_P5		; Set signal bit
		skip_bit:
			ret
	}
}

static __declspec(naked) VOID OutSigPort1Pin6(VOID)
{
	__asm {
			xor		al, ah							; Negated ^ Activated
			jz		skip_bit						; Skip Sig Bit, it is already 0
			or		dword ptr [edi], OUT_PP1_P6		; Set signal bit
		skip_bit:
			ret
	}
}

static __declspec(naked) VOID OutSigPort1Pin7(VOID)
{
	__asm {
			xor		al, ah							; Negated ^ Activated
			jz		skip_bit						; Skip Sig Bit, it is already 0
			or		dword ptr [edi], OUT_PP1_P7		; Set signal bit
		skip_bit:
			ret
	}
}

static __declspec(naked) VOID OutSigPort1Pin8(VOID)
{
	__asm {
			xor		al, ah							; Negated ^ Activated
			jz		skip_bit						; Skip Sig Bit, it is already 0
			or		dword ptr [edi], OUT_PP1_P8		; Set signal bit
		skip_bit:
			ret
	}
}

static __declspec(naked) VOID OutSigPort1Pin9(VOID)
{
	__asm {
			xor		al, ah							; Negated ^ Activated
			jz		skip_bit						; Skip Sig Bit, it is already 0
			or		dword ptr [edi], OUT_PP1_P9		; Set signal bit
		skip_bit:
			ret
	}
}

static __declspec(naked) VOID OutSigPort1Pin14(VOID)
{
	__asm {
			xor		al, ah							; Negated ^ Activated
			jz		skip_bit						; Skip Sig Bit, it is already 0
			or		dword ptr [edi], OUT_PP1_P14	; Set signal bit
		skip_bit:
			ret
	}
}

static __declspec(naked) VOID OutSigPort1Pin16(VOID)
{
	__asm {
			xor		al, ah							; Negated ^ Activated
			jz		skip_bit						; Skip Sig Bit, it is already 0
			or		dword ptr [edi], OUT_PP1_P16	; Set signal bit
		skip_bit:
			ret
	}
}

static __declspec(naked) VOID OutSigPort1Pin17(VOID)
{
	__asm {
			xor		al, ah							; Negated ^ Activated
			jz		skip_bit						; Skip Sig Bit, it is already 0
			or		dword ptr [edi], OUT_PP1_P17	; Set signal bit
		skip_bit:
			ret
	}
}

static __declspec(naked) VOID OutSigPort2Pin1(VOID)
{
	__asm {
			xor		al, ah							; Negated ^ Activated
			jz		skip_bit						; Skip Sig Bit, it is already 0
			or		dword ptr [edi], OUT_PP2_P1		; Set signal bit
		skip_bit:
			ret
	}
}

static __declspec(naked) VOID OutSigPort2Pin2(VOID)
{
	__asm {
			xor		al, ah							; Negated ^ Activated
			jz		skip_bit						; Skip Sig Bit, it is already 0
			or		dword ptr [edi], OUT_PP2_P2		; Set signal bit
		skip_bit:
			ret
	}
}

static __declspec(naked) VOID OutSigPort2Pin3(VOID)
{
	__asm {
			xor		al, ah							; Negated ^ Activated
			jz		skip_bit						; Skip Sig Bit, it is already 0
			or		dword ptr [edi], OUT_PP2_P3		; Set signal bit
		skip_bit:
			ret
	}
}

static __declspec(naked) VOID OutSigPort2Pin4(VOID)
{
	__asm {
			xor		al, ah							; Negated ^ Activated
			jz		skip_bit						; Skip Sig Bit, it is already 0
			or		dword ptr [edi], OUT_PP2_P4		; Set signal bit
		skip_bit:
			ret
	}
}

static __declspec(naked) VOID OutSigPort2Pin5(VOID)
{
	__asm {
			xor		al, ah							; Negated ^ Activated
			jz		skip_bit						; Skip Sig Bit, it is already 0
			or		dword ptr [edi], OUT_PP2_P5		; Set signal bit
		skip_bit:
			ret
	}
}

static __declspec(naked) VOID OutSigPort2Pin6(VOID)
{
	__asm {
			xor		al, ah							; Negated ^ Activated
			jz		skip_bit						; Skip Sig Bit, it is already 0
			or		dword ptr [edi], OUT_PP2_P6		; Set signal bit
		skip_bit:
			ret
	}
}

static __declspec(naked) VOID OutSigPort2Pin7(VOID)
{
	__asm {
			xor		al, ah							; Negated ^ Activated
			jz		skip_bit						; Skip Sig Bit, it is already 0
			or		dword ptr [edi], OUT_PP2_P7		; Set signal bit
		skip_bit:
			ret
	}
}

static __declspec(naked) VOID OutSigPort2Pin8(VOID)
{
	__asm {
			xor		al, ah							; Negated ^ Activated
			jz		skip_bit						; Skip Sig Bit, it is already 0
			or		dword ptr [edi], OUT_PP2_P8		; Set signal bit
		skip_bit:
			ret
	}
}

static __declspec(naked) VOID OutSigPort2Pin9(VOID)
{
	__asm {
			xor		al, ah							; Negated ^ Activated
			jz		skip_bit						; Skip Sig Bit, it is already 0
			or		dword ptr [edi], OUT_PP2_P9		; Set signal bit
		skip_bit:
			ret
	}
}

static __declspec(naked) VOID OutSigPort2Pin14(VOID)
{
	__asm {
			xor		al, ah							; Negated ^ Activated
			jz		skip_bit						; Skip Sig Bit, it is already 0
			or		dword ptr [edi], OUT_PP2_P14	; Set signal bit
		skip_bit:
			ret
	}
}

static __declspec(naked) VOID OutSigPort2Pin16(VOID)
{
	__asm {
			xor		al, ah							; Negated ^ Activated
			jz		skip_bit						; Skip Sig Bit, it is already 0
			or		dword ptr [edi], OUT_PP2_P16	; Set signal bit
		skip_bit:
			ret
	}
}

static __declspec(naked) VOID OutSigPort2Pin17(VOID)
{
	__asm {
			xor		al, ah							; Negated ^ Activated
			jz		skip_bit						; Skip Sig Bit, it is already 0
			or		dword ptr [edi], OUT_PP2_P17	; Set signal bit
		skip_bit:
			ret
	}
}

//
// This is exported OutSig functions map that maps signal handler functions
// to port pins. It is used to populate actual function arrays in mach3.c on
// driver initialization to have dedicated handlers for each signal without
// a need to check pin and port numbers to decide what to do on every IRQ.
//
const pPutSigFunc OutPPMap[PP_NUM_SIGNAL_PINS * 2] =
{
	OutSigPort1Pin1,
	OutSigPort1Pin2,
	OutSigPort1Pin3,
	OutSigPort1Pin4,
	OutSigPort1Pin5,
	OutSigPort1Pin6,
	OutSigPort1Pin7,
	OutSigPort1Pin8,
	OutSigPort1Pin9,
	0,
	0,
	0,
	0,
	OutSigPort1Pin14,
	0,
	OutSigPort1Pin16,
	OutSigPort1Pin17,
	OutSigPort2Pin1,
	OutSigPort2Pin2,
	OutSigPort2Pin3,
	OutSigPort2Pin4,
	OutSigPort2Pin5,
	OutSigPort2Pin6,
	OutSigPort2Pin7,
	OutSigPort2Pin8,
	OutSigPort2Pin9,
	0,
	0,
	0,
	0,
	OutSigPort2Pin14,
	0,
	OutSigPort2Pin16,
	OutSigPort2Pin17
};

/*
 * Same thing as for outputs above but kinda in reverse. All port inputs are
 * read in the DWORD with shifts, then we process them to set the InputInfo
 * structures in Mach3. Bytes layout is the same as above with Status
 * registers (inputs) instead of Control (outputs)
 */
#define IN_PP1_P2	0x00000001
#define IN_PP1_P3	0x00000002
#define IN_PP1_P4	0x00000004
#define IN_PP1_P5	0x00000008
#define IN_PP1_P6	0x00000010
#define IN_PP1_P7	0x00000020
#define IN_PP1_P8	0x00000040
#define IN_PP1_P9	0x00000080
#define IN_PP1_P10	0x00004000
#define IN_PP1_P11	0x00008000
#define IN_PP1_P12	0x00002000
#define IN_PP1_P13	0x00001000
#define IN_PP1_P15	0x00000800
#define IN_PP2_P2	0x00010000
#define IN_PP2_P3	0x00020000
#define IN_PP2_P4	0x00040000
#define IN_PP2_P5	0x00080000
#define IN_PP2_P6	0x00100000
#define IN_PP2_P7	0x00200000
#define IN_PP2_P8	0x00400000
#define IN_PP2_P9	0x00800000
#define IN_PP2_P10	0x40000000
#define IN_PP2_P11	0x80000000
#define IN_PP2_P12	0x20000000
#define IN_PP2_P13	0x10000000
#define IN_PP2_P15	0x08000000


// Bulk of Port Input Signal Handlers
static __declspec(naked) VOID InSigPort1Pin2(VOID)
{
	__asm {
			test	dword ptr [eax], IN_PP1_P2
			setnz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID nInSigPort1Pin2(VOID)
{
	__asm {
			test	dword ptr [eax], IN_PP1_P2
			setz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID InSigPort1Pin3(VOID)
{
	__asm {
			test	dword ptr [eax], IN_PP1_P3
			setnz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID nInSigPort1Pin3(VOID)
{
	__asm {
			test	dword ptr [eax], IN_PP1_P3
			setz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID InSigPort1Pin4(VOID)
{
	__asm {
			test	dword ptr [eax], IN_PP1_P4
			setnz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID nInSigPort1Pin4(VOID)
{
	__asm {
			test	dword ptr [eax], IN_PP1_P4
			setz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID InSigPort1Pin5(VOID)
{
	__asm {
			test	dword ptr [eax], IN_PP1_P5
			setnz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID nInSigPort1Pin5(VOID)
{
	__asm {
			test	dword ptr [eax], IN_PP1_P5
			setz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID InSigPort1Pin6(VOID)
{
	__asm {
			test	dword ptr [eax], IN_PP1_P6
			setnz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID nInSigPort1Pin6(VOID)
{
	__asm {
			test	dword ptr [eax], IN_PP1_P6
			setz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID InSigPort1Pin7(VOID)
{
	__asm {
			test	dword ptr [eax], IN_PP1_P7
			setnz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID nInSigPort1Pin7(VOID)
{
	__asm {
			test	dword ptr [eax], IN_PP1_P7
			setz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID InSigPort1Pin8(VOID)
{
	__asm {
			test	dword ptr [eax], IN_PP1_P8
			setnz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID nInSigPort1Pin8(VOID)
{
	__asm {
			test	dword ptr [eax], IN_PP1_P8
			setz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID InSigPort1Pin9(VOID)
{
	__asm {
			test	dword ptr [eax], IN_PP1_P9
			setnz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID nInSigPort1Pin9(VOID)
{
	__asm {
			test	dword ptr [eax], IN_PP1_P9
			setz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID InSigPort1Pin10(VOID)
{
	__asm {
			test	dword ptr [eax], IN_PP1_P10
			setnz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID nInSigPort1Pin10(VOID)
{
	__asm {
			test	dword ptr [eax], IN_PP1_P10
			setz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID InSigPort1Pin11(VOID)
{
	__asm {
			test	dword ptr [eax], IN_PP1_P11
			setnz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID nInSigPort1Pin11(VOID)
{
	__asm {
			test	dword ptr [eax], IN_PP1_P11
			setz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID InSigPort1Pin12(VOID)
{
	__asm {
			test	dword ptr [eax], IN_PP1_P12
			setnz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID nInSigPort1Pin12(VOID)
{
	__asm {
			test	dword ptr [eax], IN_PP1_P12
			setz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID InSigPort1Pin13(VOID)
{
	__asm {
			test	dword ptr [eax], IN_PP1_P13
			setnz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID nInSigPort1Pin13(VOID)
{
	__asm {
			test	dword ptr [eax], IN_PP1_P13
			setz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID InSigPort1Pin15(VOID)
{
	__asm {
			test	dword ptr [eax], IN_PP1_P15
			setnz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID nInSigPort1Pin15(VOID)
{
	__asm {
			test	dword ptr [eax], IN_PP1_P15
			setz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID InSigPort2Pin2(VOID)
{
	__asm {
			test	dword ptr [eax], IN_PP2_P2
			setnz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID nInSigPort2Pin2(VOID)
{
	__asm {
			test	dword ptr [eax], IN_PP2_P2
			setz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID InSigPort2Pin3(VOID)
{
	__asm {
			test	dword ptr [eax], IN_PP2_P3
			setnz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID nInSigPort2Pin3(VOID)
{
	__asm {
			test	dword ptr [eax], IN_PP2_P3
			setz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID InSigPort2Pin4(VOID)
{
	__asm {
			test	dword ptr [eax], IN_PP2_P4
			setnz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID nInSigPort2Pin4(VOID)
{
	__asm {
			test	dword ptr [eax], IN_PP2_P4
			setz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID InSigPort2Pin5(VOID)
{
	__asm {
			test	dword ptr [eax], IN_PP2_P5
			setnz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID nInSigPort2Pin5(VOID)
{
	__asm {
			test	dword ptr [eax], IN_PP2_P5
			setz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID InSigPort2Pin6(VOID)
{
	__asm {
			test	dword ptr [eax], IN_PP2_P6
			setnz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID nInSigPort2Pin6(VOID)
{
	__asm {
			test	dword ptr [eax], IN_PP2_P6
			setz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID InSigPort2Pin7(VOID)
{
	__asm {
			test	dword ptr [eax], IN_PP2_P7
			setnz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID nInSigPort2Pin7(VOID)
{
	__asm {
			test	dword ptr [eax], IN_PP2_P7
			setz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID InSigPort2Pin8(VOID)
{
	__asm {
			test	dword ptr [eax], IN_PP2_P8
			setnz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID nInSigPort2Pin8(VOID)
{
	__asm {
			test	dword ptr [eax], IN_PP2_P8
			setz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID InSigPort2Pin9(VOID)
{
	__asm {
			test	dword ptr [eax], IN_PP2_P9
			setnz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID nInSigPort2Pin9(VOID)
{
	__asm {
			test	dword ptr [eax], IN_PP2_P9
			setz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID InSigPort2Pin10(VOID)
{
	__asm {
			test	dword ptr [eax], IN_PP2_P10
			setnz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID nInSigPort2Pin10(VOID)
{
	__asm {
			test	dword ptr [eax], IN_PP2_P10
			setz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID InSigPort2Pin11(VOID)
{
	__asm {
			test	dword ptr [eax], IN_PP2_P11
			setnz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID nInSigPort2Pin11(VOID)
{
	__asm {
			test	dword ptr [eax], IN_PP2_P11
			setz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID InSigPort2Pin12(VOID)
{
	__asm {
			test	dword ptr [eax], IN_PP2_P12
			setnz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID nInSigPort2Pin12(VOID)
{
	__asm {
			test	dword ptr [eax], IN_PP2_P12
			setz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID InSigPort2Pin13(VOID)
{
	__asm {
			test	dword ptr [eax], IN_PP2_P13
			setnz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID nInSigPort2Pin13(VOID)
{
	__asm {
			test	dword ptr [eax], IN_PP2_P13
			setz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID InSigPort2Pin15(VOID)
{
	__asm {
			test	dword ptr [eax], IN_PP2_P15
			setnz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID nInSigPort2Pin15(VOID)
{
	__asm {
			test	dword ptr [eax], IN_PP2_P15
			setz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

//
// This is exported InSig functions map that maps signal handler functions
// to port pins. It is used to populate actual function arrays in mach3.c on
// driver initialization to have dedicated handlers for each signal without
// a need to check pin and port numbers to decide what to do on every IRQ.
// For inputs we have 2 arrays, one for straight signals and one for inverted
// (Negated) ones.
//
const pGetSigFunc InPPMap[PP_NUM_SIGNAL_PINS * 2] =
{
	0,
	InSigPort1Pin2,
	InSigPort1Pin3,
	InSigPort1Pin4,
	InSigPort1Pin5,
	InSigPort1Pin6,
	InSigPort1Pin7,
	InSigPort1Pin8,
	InSigPort1Pin9,
	InSigPort1Pin10,
	InSigPort1Pin11,
	InSigPort1Pin12,
	InSigPort1Pin13,
	0,
	InSigPort1Pin15,
	0,
	0,
	0,
	InSigPort2Pin2,
	InSigPort2Pin3,
	InSigPort2Pin4,
	InSigPort2Pin5,
	InSigPort2Pin6,
	InSigPort2Pin7,
	InSigPort2Pin8,
	InSigPort2Pin9,
	InSigPort2Pin10,
	InSigPort2Pin11,
	InSigPort2Pin12,
	InSigPort2Pin13,
	0,
	InSigPort2Pin15,
	0,
	0
};

const pGetSigFunc nInPPMap[PP_NUM_SIGNAL_PINS * 2] =
{
	0,
	nInSigPort1Pin2,
	nInSigPort1Pin3,
	nInSigPort1Pin4,
	nInSigPort1Pin5,
	nInSigPort1Pin6,
	nInSigPort1Pin7,
	nInSigPort1Pin8,
	nInSigPort1Pin9,
	nInSigPort1Pin10,
	nInSigPort1Pin11,
	nInSigPort1Pin12,
	nInSigPort1Pin13,
	0,
	nInSigPort1Pin15,
	0,
	0,
	0,
	nInSigPort2Pin2,
	nInSigPort2Pin3,
	nInSigPort2Pin4,
	nInSigPort2Pin5,
	nInSigPort2Pin6,
	nInSigPort2Pin7,
	nInSigPort2Pin8,
	nInSigPort2Pin9,
	nInSigPort2Pin10,
	nInSigPort2Pin11,
	nInSigPort2Pin12,
	nInSigPort2Pin13,
	0,
	nInSigPort2Pin15,
	0,
	0
};

//
// Bulk of Centipede Input/Output functions. Pin to bit mapping is 1:1.
// These work on its own In/Out buffers that are next to the Port ones
// in FdoData structure so address is "+4". Same masks for In/Out.
//
#define CPD_P1	0x00000001
#define CPD_P2	0x00000002
#define CPD_P3	0x00000004
#define CPD_P4	0x00000008
#define CPD_P5	0x00000010
#define CPD_P6	0x00000020
#define CPD_P7	0x00000040
#define CPD_P8	0x00000080
#define CPD_P9	0x00000100
#define CPD_P10	0x00000200
#define CPD_P11	0x00000400
#define CPD_P12	0x00000800
#define CPD_P13	0x00001000
#define CPD_P14	0x00002000
#define CPD_P15 0x00004000
#define CPD_P16 0x00008000
#define CPD_P17 0x00010000
#define CPD_P18 0x00020000
#define CPD_P19 0x00040000
#define CPD_P20 0x00080000
#define CPD_P21 0x00100000
#define CPD_P22 0x00200000
#define CPD_P23 0x00400000
#define CPD_P24 0x00800000
#define CPD_P25 0x01000000
#define CPD_P26 0x02000000
#define CPD_P27 0x04000000
#define CPD_P28 0x08000000
#define CPD_P29 0x10000000
#define CPD_P30 0x20000000
#define CPD_P31 0x40000000
#define CPD_P32 0x80000000

static __declspec(naked) VOID OutSigCPDPin1(VOID)
{
	__asm {
			xor		al, ah						; Negated ^ Activated
			jz		skip_bit					; Skip Sig Bit, it is already 0
			or		dword ptr [edi+4], CPD_P1	; Set signal bit
		skip_bit:
			ret
	}
}

static __declspec(naked) VOID InSigCPDPin1(VOID)
{
	__asm {
			test	dword ptr [eax+4], CPD_P1
			setnz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID nInSigCPDPin1(VOID)
{
	__asm {
			test	dword ptr [eax+4], CPD_P1
			setz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID OutSigCPDPin2(VOID)
{
	__asm {
			xor		al, ah						; Negated ^ Activated
			jz		skip_bit					; Skip Sig Bit, it is already 0
			or		dword ptr [edi+4], CPD_P2	; Set signal bit
		skip_bit:
			ret
	}
}

static __declspec(naked) VOID InSigCPDPin2(VOID)
{
	__asm {
			test	dword ptr [eax+4], CPD_P2
			setnz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID nInSigCPDPin2(VOID)
{
	__asm {
			test	dword ptr [eax+4], CPD_P2
			setz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID OutSigCPDPin3(VOID)
{
	__asm {
			xor		al, ah						; Negated ^ Activated
			jz		skip_bit					; Skip Sig Bit, it is already 0
			or		dword ptr [edi+4], CPD_P3	; Set signal bit
		skip_bit:
			ret
	}
}

static __declspec(naked) VOID InSigCPDPin3(VOID)
{
	__asm {
			test	dword ptr [eax+4], CPD_P3
			setnz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID nInSigCPDPin3(VOID)
{
	__asm {
			test	dword ptr [eax+4], CPD_P3
			setz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID OutSigCPDPin4(VOID)
{
	__asm {
			xor		al, ah						; Negated ^ Activated
			jz		skip_bit					; Skip Sig Bit, it is already 0
			or		dword ptr [edi+4], CPD_P4	; Set signal bit
		skip_bit:
			ret
	}
}

static __declspec(naked) VOID InSigCPDPin4(VOID)
{
	__asm {
			test	dword ptr [eax+4], CPD_P4
			setnz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID nInSigCPDPin4(VOID)
{
	__asm {
			test	dword ptr [eax+4], CPD_P4
			setz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID OutSigCPDPin5(VOID)
{
	__asm {
			xor		al, ah						; Negated ^ Activated
			jz		skip_bit					; Skip Sig Bit, it is already 0
			or		dword ptr [edi+4], CPD_P5	; Set signal bit
		skip_bit:
			ret
	}
}

static __declspec(naked) VOID InSigCPDPin5(VOID)
{
	__asm {
			test	dword ptr [eax+4], CPD_P5
			setnz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID nInSigCPDPin5(VOID)
{
	__asm {
			test	dword ptr [eax+4], CPD_P5
			setz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID OutSigCPDPin6(VOID)
{
	__asm {
			xor		al, ah						; Negated ^ Activated
			jz		skip_bit					; Skip Sig Bit, it is already 0
			or		dword ptr [edi+4], CPD_P6	; Set signal bit
		skip_bit:
			ret
	}
}

static __declspec(naked) VOID InSigCPDPin6(VOID)
{
	__asm {
			test	dword ptr [eax+4], CPD_P6
			setnz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID nInSigCPDPin6(VOID)
{
	__asm {
			test	dword ptr [eax+4], CPD_P6
			setz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID OutSigCPDPin7(VOID)
{
	__asm {
			xor		al, ah						; Negated ^ Activated
			jz		skip_bit					; Skip Sig Bit, it is already 0
			or		dword ptr [edi+4], CPD_P7	; Set signal bit
		skip_bit:
			ret
	}
}

static __declspec(naked) VOID InSigCPDPin7(VOID)
{
	__asm {
			test	dword ptr [eax+4], CPD_P7
			setnz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID nInSigCPDPin7(VOID)
{
	__asm {
			test	dword ptr [eax+4], CPD_P7
			setz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID OutSigCPDPin8(VOID)
{
	__asm {
			xor		al, ah						; Negated ^ Activated
			jz		skip_bit					; Skip Sig Bit, it is already 0
			or		dword ptr [edi+4], CPD_P8	; Set signal bit
		skip_bit:
			ret
	}
}

static __declspec(naked) VOID InSigCPDPin8(VOID)
{
	__asm {
			test	dword ptr [eax+4], CPD_P8
			setnz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID nInSigCPDPin8(VOID)
{
	__asm {
			test	dword ptr [eax+4], CPD_P8
			setz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID OutSigCPDPin9(VOID)
{
	__asm {
			xor		al, ah						; Negated ^ Activated
			jz		skip_bit					; Skip Sig Bit, it is already 0
			or		dword ptr [edi+4], CPD_P9	; Set signal bit
		skip_bit:
			ret
	}
}

static __declspec(naked) VOID InSigCPDPin9(VOID)
{
	__asm {
			test	dword ptr [eax+4], CPD_P9
			setnz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID nInSigCPDPin9(VOID)
{
	__asm {
			test	dword ptr [eax+4], CPD_P9
			setz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID OutSigCPDPin10(VOID)
{
	__asm {
			xor		al, ah						; Negated ^ Activated
			jz		skip_bit					; Skip Sig Bit, it is already 0
			or		dword ptr [edi+4], CPD_P10	; Set signal bit
		skip_bit:
			ret
	}
}

static __declspec(naked) VOID InSigCPDPin10(VOID)
{
	__asm {
			test	dword ptr [eax+4], CPD_P10
			setnz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID nInSigCPDPin10(VOID)
{
	__asm {
			test	dword ptr [eax+4], CPD_P10
			setz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID OutSigCPDPin11(VOID)
{
	__asm {
			xor		al, ah						; Negated ^ Activated
			jz		skip_bit					; Skip Sig Bit, it is already 0
			or		dword ptr [edi+4], CPD_P11	; Set signal bit
		skip_bit:
			ret
	}
}

static __declspec(naked) VOID InSigCPDPin11(VOID)
{
	__asm {
			test	dword ptr [eax+4], CPD_P11
			setnz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID nInSigCPDPin11(VOID)
{
	__asm {
			test	dword ptr [eax+4], CPD_P11
			setz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID OutSigCPDPin12(VOID)
{
	__asm {
			xor		al, ah						; Negated ^ Activated
			jz		skip_bit					; Skip Sig Bit, it is already 0
			or		dword ptr [edi+4], CPD_P12	; Set signal bit
		skip_bit:
			ret
	}
}

static __declspec(naked) VOID InSigCPDPin12(VOID)
{
	__asm {
			test	dword ptr [eax+4], CPD_P12
			setnz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID nInSigCPDPin12(VOID)
{
	__asm {
			test	dword ptr [eax+4], CPD_P12
			setz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID OutSigCPDPin13(VOID)
{
	__asm {
			xor		al, ah						; Negated ^ Activated
			jz		skip_bit					; Skip Sig Bit, it is already 0
			or		dword ptr [edi+4], CPD_P13	; Set signal bit
		skip_bit:
			ret
	}
}

static __declspec(naked) VOID InSigCPDPin13(VOID)
{
	__asm {
			test	dword ptr [eax+4], CPD_P13
			setnz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID nInSigCPDPin13(VOID)
{
	__asm {
			test	dword ptr [eax+4], CPD_P13
			setz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID OutSigCPDPin14(VOID)
{
	__asm {
			xor		al, ah						; Negated ^ Activated
			jz		skip_bit					; Skip Sig Bit, it is already 0
			or		dword ptr [edi+4], CPD_P14	; Set signal bit
		skip_bit:
			ret
	}
}

static __declspec(naked) VOID InSigCPDPin14(VOID)
{
	__asm {
			test	dword ptr [eax+4], CPD_P14
			setnz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID nInSigCPDPin14(VOID)
{
	__asm {
			test	dword ptr [eax+4], CPD_P14
			setz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID OutSigCPDPin15(VOID)
{
	__asm {
			xor		al, ah						; Negated ^ Activated
			jz		skip_bit					; Skip Sig Bit, it is already 0
			or		dword ptr [edi+4], CPD_P15	; Set signal bit
		skip_bit:
			ret
	}
}

static __declspec(naked) VOID InSigCPDPin15(VOID)
{
	__asm {
			test	dword ptr [eax+4], CPD_P15
			setnz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID nInSigCPDPin15(VOID)
{
	__asm {
			test	dword ptr [eax+4], CPD_P15
			setz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID OutSigCPDPin16(VOID)
{
	__asm {
			xor		al, ah						; Negated ^ Activated
			jz		skip_bit					; Skip Sig Bit, it is already 0
			or		dword ptr [edi+4], CPD_P16	; Set signal bit
		skip_bit:
			ret
	}
}

static __declspec(naked) VOID InSigCPDPin16(VOID)
{
	__asm {
			test	dword ptr [eax+4], CPD_P16
			setnz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID nInSigCPDPin16(VOID)
{
	__asm {
			test	dword ptr [eax+4], CPD_P16
			setz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID OutSigCPDPin17(VOID)
{
	__asm {
			xor		al, ah						; Negated ^ Activated
			jz		skip_bit					; Skip Sig Bit, it is already 0
			or		dword ptr [edi+4], CPD_P17	; Set signal bit
		skip_bit:
			ret
	}
}

static __declspec(naked) VOID InSigCPDPin17(VOID)
{
	__asm {
			test	dword ptr [eax+4], CPD_P17
			setnz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID nInSigCPDPin17(VOID)
{
	__asm {
			test	dword ptr [eax+4], CPD_P17
			setz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID OutSigCPDPin18(VOID)
{
	__asm {
			xor		al, ah						; Negated ^ Activated
			jz		skip_bit					; Skip Sig Bit, it is already 0
			or		dword ptr [edi+4], CPD_P18	; Set signal bit
		skip_bit:
			ret
	}
}

static __declspec(naked) VOID InSigCPDPin18(VOID)
{
	__asm {
			test	dword ptr [eax+4], CPD_P18
			setnz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID nInSigCPDPin18(VOID)
{
	__asm {
			test	dword ptr [eax+4], CPD_P18
			setz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID OutSigCPDPin19(VOID)
{
	__asm {
			xor		al, ah						; Negated ^ Activated
			jz		skip_bit					; Skip Sig Bit, it is already 0
			or		dword ptr [edi+4], CPD_P19	; Set signal bit
		skip_bit:
			ret
	}
}

static __declspec(naked) VOID InSigCPDPin19(VOID)
{
	__asm {
			test	dword ptr [eax+4], CPD_P19
			setnz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID nInSigCPDPin19(VOID)
{
	__asm {
			test	dword ptr [eax+4], CPD_P19
			setz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID OutSigCPDPin20(VOID)
{
	__asm {
			xor		al, ah						; Negated ^ Activated
			jz		skip_bit					; Skip Sig Bit, it is already 0
			or		dword ptr [edi+4], CPD_P20	; Set signal bit
		skip_bit:
			ret
	}
}

static __declspec(naked) VOID InSigCPDPin20(VOID)
{
	__asm {
			test	dword ptr [eax+4], CPD_P20
			setnz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID nInSigCPDPin20(VOID)
{
	__asm {
			test	dword ptr [eax+4], CPD_P20
			setz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID OutSigCPDPin21(VOID)
{
	__asm {
			xor		al, ah						; Negated ^ Activated
			jz		skip_bit					; Skip Sig Bit, it is already 0
			or		dword ptr [edi+4], CPD_P21	; Set signal bit
		skip_bit:
			ret
	}
}

static __declspec(naked) VOID InSigCPDPin21(VOID)
{
	__asm {
			test	dword ptr [eax+4], CPD_P21
			setnz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID nInSigCPDPin21(VOID)
{
	__asm {
			test	dword ptr [eax+4], CPD_P21
			setz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID OutSigCPDPin22(VOID)
{
	__asm {
			xor		al, ah						; Negated ^ Activated
			jz		skip_bit					; Skip Sig Bit, it is already 0
			or		dword ptr [edi+4], CPD_P22	; Set signal bit
		skip_bit:
			ret
	}
}

static __declspec(naked) VOID InSigCPDPin22(VOID)
{
	__asm {
			test	dword ptr [eax+4], CPD_P22
			setnz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID nInSigCPDPin22(VOID)
{
	__asm {
			test	dword ptr [eax+4], CPD_P22
			setz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID OutSigCPDPin23(VOID)
{
	__asm {
			xor		al, ah						; Negated ^ Activated
			jz		skip_bit					; Skip Sig Bit, it is already 0
			or		dword ptr [edi+4], CPD_P23	; Set signal bit
		skip_bit:
			ret
	}
}

static __declspec(naked) VOID InSigCPDPin23(VOID)
{
	__asm {
			test	dword ptr [eax+4], CPD_P23
			setnz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID nInSigCPDPin23(VOID)
{
	__asm {
			test	dword ptr [eax+4], CPD_P23
			setz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID OutSigCPDPin24(VOID)
{
	__asm {
			xor		al, ah						; Negated ^ Activated
			jz		skip_bit					; Skip Sig Bit, it is already 0
			or		dword ptr [edi+4], CPD_P24	; Set signal bit
		skip_bit:
			ret
	}
}

static __declspec(naked) VOID InSigCPDPin24(VOID)
{
	__asm {
			test	dword ptr [eax+4], CPD_P24
			setnz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID nInSigCPDPin24(VOID)
{
	__asm {
			test	dword ptr [eax+4], CPD_P24
			setz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID OutSigCPDPin25(VOID)
{
	__asm {
			xor		al, ah						; Negated ^ Activated
			jz		skip_bit					; Skip Sig Bit, it is already 0
			or		dword ptr [edi+4], CPD_P25	; Set signal bit
		skip_bit:
			ret
	}
}

static __declspec(naked) VOID InSigCPDPin25(VOID)
{
	__asm {
			test	dword ptr [eax+4], CPD_P25
			setnz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID nInSigCPDPin25(VOID)
{
	__asm {
			test	dword ptr [eax+4], CPD_P25
			setz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID OutSigCPDPin26(VOID)
{
	__asm {
			xor		al, ah						; Negated ^ Activated
			jz		skip_bit					; Skip Sig Bit, it is already 0
			or		dword ptr [edi+4], CPD_P26	; Set signal bit
		skip_bit:
			ret
	}
}

static __declspec(naked) VOID InSigCPDPin26(VOID)
{
	__asm {
			test	dword ptr [eax+4], CPD_P26
			setnz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID nInSigCPDPin26(VOID)
{
	__asm {
			test	dword ptr [eax+4], CPD_P26
			setz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID OutSigCPDPin27(VOID)
{
	__asm {
			xor		al, ah						; Negated ^ Activated
			jz		skip_bit					; Skip Sig Bit, it is already 0
			or		dword ptr [edi+4], CPD_P27	; Set signal bit
		skip_bit:
			ret
	}
}

static __declspec(naked) VOID InSigCPDPin27(VOID)
{
	__asm {
			test	dword ptr [eax+4], CPD_P27
			setnz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID nInSigCPDPin27(VOID)
{
	__asm {
			test	dword ptr [eax+4], CPD_P27
			setz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID OutSigCPDPin28(VOID)
{
	__asm {
			xor		al, ah						; Negated ^ Activated
			jz		skip_bit					; Skip Sig Bit, it is already 0
			or		dword ptr [edi+4], CPD_P28	; Set signal bit
		skip_bit:
			ret
	}
}

static __declspec(naked) VOID InSigCPDPin28(VOID)
{
	__asm {
			test	dword ptr [eax+4], CPD_P28
			setnz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID nInSigCPDPin28(VOID)
{
	__asm {
			test	dword ptr [eax+4], CPD_P28
			setz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID OutSigCPDPin29(VOID)
{
	__asm {
			xor		al, ah						; Negated ^ Activated
			jz		skip_bit					; Skip Sig Bit, it is already 0
			or		dword ptr [edi+4], CPD_P29	; Set signal bit
		skip_bit:
			ret
	}
}

static __declspec(naked) VOID InSigCPDPin29(VOID)
{
	__asm {
			test	dword ptr [eax+4], CPD_P29
			setnz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID nInSigCPDPin29(VOID)
{
	__asm {
			test	dword ptr [eax+4], CPD_P29
			setz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID OutSigCPDPin30(VOID)
{
	__asm {
			xor		al, ah						; Negated ^ Activated
			jz		skip_bit					; Skip Sig Bit, it is already 0
			or		dword ptr [edi+4], CPD_P30	; Set signal bit
		skip_bit:
			ret
	}
}

static __declspec(naked) VOID InSigCPDPin30(VOID)
{
	__asm {
			test	dword ptr [eax+4], CPD_P30
			setnz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID nInSigCPDPin30(VOID)
{
	__asm {
			test	dword ptr [eax+4], CPD_P30
			setz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID OutSigCPDPin31(VOID)
{
	__asm {
			xor		al, ah						; Negated ^ Activated
			jz		skip_bit					; Skip Sig Bit, it is already 0
			or		dword ptr [edi+4], CPD_P31	; Set signal bit
		skip_bit:
			ret
	}
}

static __declspec(naked) VOID InSigCPDPin31(VOID)
{
	__asm {
			test	dword ptr [eax+4], CPD_P31
			setnz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID nInSigCPDPin31(VOID)
{
	__asm {
			test	dword ptr [eax+4], CPD_P31
			setz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID OutSigCPDPin32(VOID)
{
	__asm {
			xor		al, ah						; Negated ^ Activated
			jz		skip_bit					; Skip Sig Bit, it is already 0
			or		dword ptr [edi+4], CPD_P32	; Set signal bit
		skip_bit:
			ret
	}
}

static __declspec(naked) VOID InSigCPDPin32(VOID)
{
	__asm {
			test	dword ptr [eax+4], CPD_P32
			setnz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

static __declspec(naked) VOID nInSigCPDPin32(VOID)
{
	__asm {
			test	dword ptr [eax+4], CPD_P32
			setz	byte ptr [edi+8]			; Activated <- (!0) ? 1 : 0
			ret
	}
}

// Exported Centipede maps
const pPutSigFunc OutCPDMap[32] =
{
	OutSigCPDPin1,
	OutSigCPDPin2,
	OutSigCPDPin3,
	OutSigCPDPin4,
	OutSigCPDPin5,
	OutSigCPDPin6,
	OutSigCPDPin7,
	OutSigCPDPin8,
	OutSigCPDPin9,
	OutSigCPDPin10,
	OutSigCPDPin11,
	OutSigCPDPin12,
	OutSigCPDPin13,
	OutSigCPDPin14,
	OutSigCPDPin15,
	OutSigCPDPin16,
	OutSigCPDPin17,
	OutSigCPDPin18,
	OutSigCPDPin19,
	OutSigCPDPin20,
	OutSigCPDPin21,
	OutSigCPDPin22,
	OutSigCPDPin23,
	OutSigCPDPin24,
	OutSigCPDPin25,
	OutSigCPDPin26,
	OutSigCPDPin27,
	OutSigCPDPin28,
	OutSigCPDPin29,
	OutSigCPDPin30,
	OutSigCPDPin31,
	OutSigCPDPin32
};

const pGetSigFunc InCPDMap[32] =
{
	InSigCPDPin1,
	InSigCPDPin2,
	InSigCPDPin3,
	InSigCPDPin4,
	InSigCPDPin5,
	InSigCPDPin6,
	InSigCPDPin7,
	InSigCPDPin8,
	InSigCPDPin9,
	InSigCPDPin10,
	InSigCPDPin11,
	InSigCPDPin12,
	InSigCPDPin13,
	InSigCPDPin14,
	InSigCPDPin15,
	InSigCPDPin16,
	InSigCPDPin17,
	InSigCPDPin18,
	InSigCPDPin19,
	InSigCPDPin20,
	InSigCPDPin21,
	InSigCPDPin22,
	InSigCPDPin23,
	InSigCPDPin24,
	InSigCPDPin25,
	InSigCPDPin26,
	InSigCPDPin27,
	InSigCPDPin28,
	InSigCPDPin29,
	InSigCPDPin30,
	InSigCPDPin31,
	InSigCPDPin32
};

const pGetSigFunc nInCPDMap[32] =
{
	nInSigCPDPin1,
	nInSigCPDPin2,
	nInSigCPDPin3,
	nInSigCPDPin4,
	nInSigCPDPin5,
	nInSigCPDPin6,
	nInSigCPDPin7,
	nInSigCPDPin8,
	nInSigCPDPin9,
	nInSigCPDPin10,
	nInSigCPDPin11,
	nInSigCPDPin12,
	nInSigCPDPin13,
	nInSigCPDPin14,
	nInSigCPDPin15,
	nInSigCPDPin16,
	nInSigCPDPin17,
	nInSigCPDPin18,
	nInSigCPDPin19,
	nInSigCPDPin20,
	nInSigCPDPin21,
	nInSigCPDPin22,
	nInSigCPDPin23,
	nInSigCPDPin24,
	nInSigCPDPin25,
	nInSigCPDPin26,
	nInSigCPDPin27,
	nInSigCPDPin28,
	nInSigCPDPin29,
	nInSigCPDPin30,
	nInSigCPDPin31,
	nInSigCPDPin32
};


// ***********************************************************************
// * Utility function, maps one Port/Pin to a GetSig() function pointer. *
// ***********************************************************************
VOID MapOneInput(PFDO_DATA FdoData, pGetSigFunc *GetSigMap, INT Port,
			INT Pin, BOOLEAN Negated)
{
	switch (Port) {
		//
		// Port 1 in Input mode when InPort is 5, regular mode if it is 1
		//
		case (PPORT1 | PP_INPUT):
			if (!FdoData->Port1isOK) goto BadPPIn;
			if (!FdoData->Port1On) {
				FdoData->Port1On = TRUE;
				FdoData->Port1isInput = TRUE;
			} else if (!FdoData->Port1isInput) {
				goto BadPPIn;
			}
			goto MapPPIn1;
		case PPORT1:
			if (!FdoData->Port1isOK) goto BadPPIn;
			if (!FdoData->Port1On) {
				FdoData->Port1On = TRUE;
				FdoData->Port1isInput = FALSE;
			} else if (FdoData->Port1isInput) {
				goto BadPPIn;
			}
MapPPIn1:
			if (Pin < 1 || Pin > 17) {
				*GetSigMap = NULL;
			} else {
				*GetSigMap = (Negated) ? nInPPMap[Pin-1] : InPPMap[Pin-1];
			}
			break;

		//
		// Port 2 in Input mode when InPort is 6, regular mode if it is 2
		//
		case (PPORT2 | PP_INPUT):
			if (!FdoData->Port2isOK) goto BadPPIn;
			if (!FdoData->Port2On) {
				FdoData->Port2On = TRUE;
				FdoData->Port2isInput = TRUE;
			} else if (!FdoData->Port2isInput) {
				goto BadPPIn;
			}
			goto MapPPIn2;
		case PPORT2:
			if (!FdoData->Port2isOK) goto BadPPIn;
			if (!FdoData->Port2On) {
				FdoData->Port2On = TRUE;
				FdoData->Port2isInput = FALSE;
			} else if (FdoData->Port2isInput) {
				goto BadPPIn;
			}
MapPPIn2:
			if (Pin < 1 || Pin > 17) {
				*GetSigMap = NULL;
			} else {
				*GetSigMap = (Negated) ? nInPPMap[(Pin-1)*2] : InPPMap[(Pin-1)*2];
			}
			break;
BadPPIn:
			FdoData->PortsMapOK = FALSE;
			*GetSigMap = NULL;
			break;

		//
		// CENTIPEDE ports. There is no modes for those.
		//
		case CPD_PORT:
			if (Pin < 1 || Pin > 32) {
				*GetSigMap = NULL;
			} else {
				*GetSigMap = (Negated) ? nInCPDMap[Pin-1] : InCPDMap[Pin-1];
			}
			break;

		default:
			*GetSigMap = NULL;
	}
}


// ***********************************************************************
// * Utility function, maps one Port/Pin to a PutSig() function pointer. *
// ***********************************************************************
VOID MapOneOutput(PFDO_DATA FdoData, pPutSigFunc *PutSigMap, INT Port, INT Pin)
{
	switch (Port) {

		case (PPORT1 | PP_INPUT):
			if (!FdoData->Port1isOK) goto BadPPOut;
			if (!FdoData->Port1On) {
				FdoData->Port1On = TRUE;
				FdoData->Port1isInput = TRUE;
			} else if (!FdoData->Port1isInput) {
				goto BadPPOut;
			}
			goto MapPPOut1;
		case PPORT1:
			if (!FdoData->Port1isOK) goto BadPPOut;
			if (!FdoData->Port1On) {
				FdoData->Port1On = TRUE;
				FdoData->Port1isInput = FALSE;
			} else if (FdoData->Port1isInput) {
				goto BadPPOut;
			}
MapPPOut1:
			if (Pin < 1 || Pin > 17) {
				*PutSigMap = NULL;
			} else {
				*PutSigMap = OutPPMap[Pin-1];
			}
			break;

		case (PPORT2 | PP_INPUT):
			if (!FdoData->Port2isOK) goto BadPPOut;
			if (!FdoData->Port2On) {
				FdoData->Port2On = TRUE;
				FdoData->Port2isInput = TRUE;
			} else if (!FdoData->Port2isInput) {
				goto BadPPOut;
			}
			goto MapPPOut2;
		case PPORT2:
			if (!FdoData->Port2isOK) goto BadPPOut;
			if (!FdoData->Port2On) {
				FdoData->Port2On = TRUE;
				FdoData->Port2isInput = FALSE;
			} else if (FdoData->Port2isInput) {
				goto BadPPOut;
			}
MapPPOut2:
			if (Pin < 1 || Pin > 17) {
				*PutSigMap = NULL;
			} else {
				*PutSigMap = OutPPMap[(Pin-1)*2];
			}
			break;
BadPPOut:
			FdoData->PortsMapOK = FALSE;
			*PutSigMap = NULL;
			break;

		case CPD_PORT:
			if (Pin < 1 || Pin > 32) {
				*PutSigMap = NULL;
			} else {
				*PutSigMap = OutCPDMap[Pin-1];
				FdoData->CPDOutMask &= ~(1 << (Pin-1));
			}
			break;

		default:
			*PutSigMap = NULL;
	}
}
